"""Tests for the MolToMACCSFP pipeline element."""

from __future__ import annotations

import unittest
from typing import Any

import numpy as np

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToMACCSFP

# pylint: disable=duplicate-code
# Similar to test_mol2morgan_fingerprint.py and test_mol2path_fingerprint.py

test_smiles = [
    "c1ccccc1",
    "c1ccccc1C",
    "NCCOCCCC(=O)O",
]


class TestMolToMACCSFP(unittest.TestCase):
    """Unittest for MolToMACCSFP, which calculates MACCS Key Fingerprints."""

    def test_can_be_constructed(self) -> None:
        """Test if the MolToMACCSFP pipeline element can be constructed.

        Returns
        -------
        None
        """
        mol_fp = MolToMACCSFP()
        mol_fp_copy = mol_fp.copy()
        self.assertTrue(mol_fp_copy is not mol_fp)
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_copy.get_params()[key])
        mol_fp_recreated = MolToMACCSFP(**mol_fp.get_params())
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_recreated.get_params()[key])

    def test_output_types(self) -> None:
        """Test equality of different output_types."""

        smi2mol = SmilesToMol()
        sparse_maccs = MolToMACCSFP(return_as="sparse")
        dense_maccs = MolToMACCSFP(return_as="dense")
        explicit_bit_vect_maccs = MolToMACCSFP(return_as="explicit_bit_vect")
        sparse_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("sparse_maccs", sparse_maccs),
            ],
        )
        dense_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("dense_maccs", dense_maccs),
            ],
        )
        explicit_bit_vect_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("explicit_bit_vect_maccs", explicit_bit_vect_maccs),
            ],
        )

        sparse_output = sparse_pipeline.fit_transform(test_smiles)
        dense_output = dense_pipeline.fit_transform(test_smiles)
        explicit_bit_vect_maccs_output = explicit_bit_vect_pipeline.fit_transform(
            test_smiles
        )

        self.assertTrue(np.all(sparse_output.toarray() == dense_output))

        self.assertTrue(
            np.equal(
                dense_output,
                np.array(explicit_bit_vect_maccs_output),
            ).all()
        )

    def test_setter_getter(self) -> None:
        """Test if the setters and getters work as expected."""
        mol_fp = MolToMACCSFP()
        params: dict[str, Any] = {
            "return_as": "dense",
        }
        mol_fp.set_params(**params)
        self.assertEqual(mol_fp.get_params()["return_as"], "dense")

    def test_setter_getter_error_handling(self) -> None:
        """Test if the setters and getters work as expected when errors are encountered."""

        mol_fp = MolToMACCSFP()
        params: dict[str, Any] = {
            "return_as": "invalid-option",
        }
        self.assertRaises(ValueError, mol_fp.set_params, **params)

    def test_feature_names(self) -> None:
        """Test if the feature names are correct."""
        mol_fp = MolToMACCSFP()
        feature_names = mol_fp.feature_names
        self.assertEqual(len(feature_names), mol_fp.n_bits)
        # feature names should be unique
        self.assertEqual(len(feature_names), len(set(feature_names)))


if __name__ == "__main__":
    unittest.main()



"""Test the custom filter element."""

import unittest

from molpipeline import Pipeline
from molpipeline.any2mol import AutoToMol
from molpipeline.experimental import CustomFilter
from molpipeline.mol2any import MolToBool


class TestCustomFilter(unittest.TestCase):
    """Test the custom filter element."""

    smiles_list = [
        "CC",
        "CCC",
        "CCCC",
        "CO",
    ]

    def test_transform(self) -> None:
        """Test the custom filter."""
        mol_list = AutoToMol().transform(self.smiles_list)
        res_filter = CustomFilter(lambda x: x.GetNumAtoms() == 2).transform(mol_list)
        res_bool = MolToBool().transform(res_filter)
        self.assertEqual(res_bool, [True, False, False, True])

    def test_pipeline(self) -> None:
        """Test the custom filter in pipeline."""
        pipeline = Pipeline(
            [
                ("auto_to_mol", AutoToMol()),
                ("custom_filter", CustomFilter(lambda x: x.GetNumAtoms() == 2)),
                ("mol_to_bool", MolToBool()),
            ]
        )
        self.assertEqual(
            pipeline.transform(self.smiles_list), [True, False, False, True]
        )



"""Test the comparison functions."""

from typing import Callable
from unittest import TestCase

from molpipeline import Pipeline
from molpipeline.utils.comparison import check_pipelines_equivalent
from tests.utils.default_models import (
    get_morgan_physchem_rf_pipeline,
    get_standardization_pipeline,
)


class TestComparison(TestCase):
    """Test if functional equivalent pipelines are detected as such."""

    def test_are_equal(self) -> None:
        """Test if two equivalent pipelines are detected as such."""
        # Test standardization pipelines
        pipline_method_list: list[Callable[[int], Pipeline]] = [
            get_standardization_pipeline,
            get_morgan_physchem_rf_pipeline,
        ]
        for pipeline_method in pipline_method_list:
            pipeline_a = pipeline_method(1)
            pipeline_b = pipeline_method(1)
            self.assertTrue(check_pipelines_equivalent(pipeline_a, pipeline_b))

    def test_are_not_equal(self) -> None:
        """Test if two different pipelines are detected as such."""
        # Test changed parameters
        pipeline_a = get_morgan_physchem_rf_pipeline()
        pipeline_b = get_morgan_physchem_rf_pipeline()
        pipeline_b.set_params(mol2fp__morgan__n_bits=1024)
        self.assertFalse(check_pipelines_equivalent(pipeline_a, pipeline_b))

        # Test changed steps
        pipeline_b = get_morgan_physchem_rf_pipeline()
        last_step = pipeline_b.steps[-1]
        pipeline_b.steps = pipeline_b.steps[:-1]
        self.assertFalse(check_pipelines_equivalent(pipeline_a, pipeline_b))

        # Test if adding the step back makes the pipelines equivalent
        pipeline_b.steps.append(last_step)
        self.assertTrue(check_pipelines_equivalent(pipeline_a, pipeline_b))



"""Module for custom filter functionality."""

from __future__ import annotations

from typing import Any, Callable, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
)
from molpipeline.abstract_pipeline_elements.core import (
    MolToMolPipelineElement as _MolToMolPipelineElement,
)
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol


class CustomFilter(_MolToMolPipelineElement):
    """Filters molecules based on a custom boolean function. Elements not passing the filter will be set to InvalidInstances."""

    def __init__(
        self,
        func: Callable[[RDKitMol], bool],
        name: str = "CustomFilter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize CustomFilter.

        Parameters
        ----------
        func : Callable[[RDKitMol], bool]
            custom function to filter molecules
        name : str, optional
            name of the element, by default "CustomFilter"
        n_jobs : int, optional
            number of jobs to use, by default 1
        uuid : str, optional
            uuid of the element, by default None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.func = func

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Pretransform single value.

        Applies the custom boolean function to the molecule.

        Parameters
        ----------
        value : RDKitMol
            input value

        Returns
        -------
        OptionalMol
            output value
        """
        if self.func(value):
            return value
        return InvalidInstance(
            self.uuid,
            f"Molecule does not match filter from {self.name}",
            self.name,
        )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of CustomFilter.

        Parameters
        ----------
        deep: bool, optional (default: True)
            If True, return the parameters of all subobjects that are PipelineElements.

        Returns
        -------
        dict[str, Any]
            Parameters of CustomFilter.
        """
        params = super().get_params(deep=deep)
        if deep:
            params["func"] = self.func
        else:
            params["func"] = self.func
        return params

    def set_params(self, **parameters: dict[str, Any]) -> Self:
        """Set parameters of CustomFilter.

        Parameters
        ----------
        parameters: dict[str, Any]
            Parameters to set.

        Returns
        -------
        Self
            Self.
        """
        parameter_copy = dict(parameters)
        if "func" in parameter_copy:
            self.func = parameter_copy.pop("func")  # type: ignore
        super().set_params(**parameter_copy)
        return self



"""Test SubpipelineExtractor."""

import unittest

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToMorganFP, MolToSmiles
from molpipeline.mol2mol import SaltRemover
from molpipeline.utils.subpipeline import SubpipelineExtractor


class TestSubpipelineExtractor(unittest.TestCase):
    """Test SubpipelineExtractor."""

    def test_get_molecule_reader_element(self) -> None:
        """Test extracting molecule reader element from pipelines."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_molecule_reader_element(), pipeline.steps[0][1])

        # test with multiple molecule readers
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2smi", MolToSmiles()),
                ("smi2mol2", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_molecule_reader_element(), pipeline.steps[2][1])

    def test_get_featurization_element(self) -> None:
        """Test extracting featurization element from pipelines."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_featurization_element(), pipeline.steps[1][1])

        # test with PostPredictionWrapper
        error_filter = ErrorFilter()
        error_reinserter = PostPredictionWrapper(
            FilterReinserter.from_error_filter(error_filter, None)
        )
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
                (
                    "error_reinserter",
                    error_reinserter,
                ),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_featurization_element(), pipeline.steps[2][1])

    def test_get_model_element(self) -> None:
        """Test extracting model element from pipeline."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_model_element(), pipeline.steps[2][1])

        # test with PostPredictionWrapper
        error_filter = ErrorFilter()
        error_reinserter = PostPredictionWrapper(
            FilterReinserter.from_error_filter(error_filter, None)
        )
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
                (
                    "error_reinserter",
                    error_reinserter,
                ),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        self.assertIs(extractor.get_model_element(), pipeline.steps[3][1])

    def test_get_molecule_reader_subpipeline(self) -> None:
        """Test extracting subpipeline up to the molecule reader element from pipelines."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_molecule_reader_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 1)  # type: ignore[union-attr]
        self.assertIs(subpipeline.steps[0], pipeline.steps[0])  # type: ignore[union-attr]

        # test with multiple molecule readers
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("mol2smi", MolToSmiles()),
                ("smi2mol2", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_molecule_reader_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 3)  # type: ignore[union-attr]
        for i, subpipe_step in enumerate(subpipeline.steps):  # type: ignore[union-attr]
            self.assertIs(subpipe_step, pipeline.steps[i])

    def test_get_featurization_subpipeline(self) -> None:
        """Test extracting subpipeline up to the featurization element from pipelines."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_featurization_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 2)  # type: ignore[union-attr]
        for i, subpipe_step in enumerate(subpipeline.steps):  # type: ignore[union-attr]
            self.assertIs(subpipe_step, pipeline.steps[i])

        # test with PostPredictionWrapper
        error_filter = ErrorFilter()
        error_reinserter = PostPredictionWrapper(
            FilterReinserter.from_error_filter(error_filter, None)
        )
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
                (
                    "error_reinserter",
                    error_reinserter,
                ),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_featurization_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 3)  # type: ignore[union-attr]
        for i, subpipe_step in enumerate(subpipeline.steps):  # type: ignore[union-attr]
            self.assertIs(subpipe_step, pipeline.steps[i])

    def test_get_model_subpipeline(self) -> None:
        """Test extracting subpipeline up to the model element from pipelines."""

        # test basic example
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_model_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 3)  # type: ignore[union-attr]
        for i, subpipe_step in enumerate(subpipeline.steps):  # type: ignore[union-attr]
            self.assertIs(subpipe_step, pipeline.steps[i])

        # test with PostPredictionWrapper
        error_filter = ErrorFilter()
        error_reinserter = PostPredictionWrapper(
            FilterReinserter.from_error_filter(error_filter, None)
        )
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
                (
                    "error_reinserter",
                    error_reinserter,
                ),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        subpipeline = extractor.get_model_subpipeline()
        self.assertIsInstance(subpipeline, Pipeline)
        self.assertEqual(len(subpipeline.steps), 4)  # type: ignore[union-attr]
        for i, subpipe_step in enumerate(subpipeline.steps):  # type: ignore[union-attr]
            self.assertIs(subpipe_step, pipeline.steps[i])

    def test_get_subpipeline(self) -> None:
        """Test extracting subpipeline as a certain interval from the original pipeline."""

        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                ("salt_remover", SaltRemover()),
                ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                ("model", RandomForestClassifier()),
            ]
        )
        extractor = SubpipelineExtractor(pipeline)
        reader_element = extractor.get_molecule_reader_element("smi2mol")
        self.assertIs(reader_element, pipeline.steps[0][1])
        feature_element = extractor.get_featurization_element("morgan")
        self.assertIs(feature_element, pipeline.steps[2][1])
        model_element = extractor.get_model_element("model")
        self.assertIs(model_element, pipeline.steps[3][1])

        # test smi2mol to morgan
        subpipeline_reader_feature = extractor.get_subpipeline(
            reader_element, feature_element
        )
        self.assertIsInstance(subpipeline_reader_feature, Pipeline)
        self.assertEqual(len(subpipeline_reader_feature.steps), 3)  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_feature.steps[0], pipeline.steps[0])  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_feature.steps[1], pipeline.steps[1])  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_feature.steps[2], pipeline.steps[2])  # type: ignore[union-attr]

        # test smi2mol to model
        subpipeline_reader_model = extractor.get_subpipeline(
            reader_element, model_element
        )
        self.assertIsInstance(subpipeline_reader_model, Pipeline)
        self.assertEqual(len(subpipeline_reader_model.steps), 4)  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_model.steps[0], pipeline.steps[0])  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_model.steps[1], pipeline.steps[1])  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_model.steps[2], pipeline.steps[2])  # type: ignore[union-attr]
        self.assertIs(subpipeline_reader_model.steps[3], pipeline.steps[3])  # type: ignore[union-attr]

        # test morgan to model
        subpipeline_feature_model = extractor.get_subpipeline(
            feature_element, model_element
        )
        self.assertIsInstance(subpipeline_feature_model, Pipeline)
        self.assertEqual(len(subpipeline_feature_model.steps), 2)  # type: ignore[union-attr]
        self.assertIs(subpipeline_feature_model.steps[0], pipeline.steps[2])  # type: ignore[union-attr]
        self.assertIs(subpipeline_feature_model.steps[1], pipeline.steps[3])  # type: ignore[union-attr]

        # test morgan to morgan
        subpipeline_feature_feature = extractor.get_subpipeline(
            feature_element, feature_element
        )
        self.assertIsInstance(subpipeline_feature_feature, Pipeline)
        self.assertEqual(len(subpipeline_feature_feature.steps), 1)  # type: ignore[union-attr]
        self.assertIs(subpipeline_feature_feature.steps[0], pipeline.steps[2])  # type: ignore[union-attr]

        # test the first element comes after the second element
        self.assertRaises(
            ValueError,
            extractor.get_subpipeline,
            feature_element,
            reader_element,
        )

        element_not_in_pipeline = SmilesToMol()

        # test element not in pipeline raises an exception
        self.assertRaises(
            ValueError,
            extractor.get_subpipeline,
            element_not_in_pipeline,
            feature_element,
        )
        self.assertRaises(
            ValueError,
            extractor.get_subpipeline,
            reader_element,
            element_not_in_pipeline,
        )

    def test_get_all_filter_reinserter_fill_values(self) -> None:
        """Test extracting all FilterReinserter fill values from pipelines."""

        test_fill_values = [None, np.nan]

        for test_fill_value in test_fill_values:
            error_filter = ErrorFilter()
            error_reinserter = PostPredictionWrapper(
                FilterReinserter.from_error_filter(error_filter, test_fill_value)
            )
            pipeline = Pipeline(
                [
                    ("smi2mol", SmilesToMol()),
                    ("error_filter", error_filter),
                    ("morgan", MolToMorganFP(radius=1, n_bits=64)),
                    ("model", RandomForestClassifier()),
                    (
                        "error_reinserter",
                        error_reinserter,
                    ),
                ]
            )
            extractor = SubpipelineExtractor(pipeline)
            fill_values = extractor.get_all_filter_reinserter_fill_values()
            self.assertEqual(fill_values, [test_fill_value])



"""Functions for comparing chemprop models."""

from typing import Sequence
from unittest import TestCase

import torch
from chemprop.nn.loss import LossFunction
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.profilers.base import PassThroughProfiler
from sklearn.base import BaseEstimator
from torch import nn


def compare_params(
    test_case: TestCase, model_a: BaseEstimator, model_b: BaseEstimator
) -> None:
    """Compare the parameters of two models.

    Parameters
    ----------
    test_case : TestCase
        The test case for which to raise the assertion.
    model_a : BaseEstimator
        The first model.
    model_b : BaseEstimator
        The second model.
    """
    model_a_params = model_a.get_params(deep=True)
    model_b_params = model_b.get_params(deep=True)
    test_case.assertSetEqual(set(model_a_params.keys()), set(model_b_params.keys()))
    for param_name, param_a in model_a_params.items():
        param_b = model_b_params[param_name]
        test_case.assertEqual(param_a.__class__, param_b.__class__)
        if hasattr(param_a, "get_params"):
            test_case.assertTrue(hasattr(param_b, "get_params"))
            test_case.assertNotEqual(id(param_a), id(param_b))
        elif isinstance(param_a, LossFunction):
            test_case.assertEqual(
                param_a.state_dict()["task_weights"],
                param_b.state_dict()["task_weights"],
            )
            test_case.assertEqual(type(param_a), type(param_b))
        elif isinstance(param_a, (nn.Identity, Accelerator, PassThroughProfiler)):
            test_case.assertEqual(type(param_a), type(param_b))
        elif isinstance(param_a, torch.Tensor):
            test_case.assertTrue(
                torch.equal(param_a, param_b), f"Test failed for {param_name}"
            )
        elif param_name == "lightning_trainer__callbacks":
            test_case.assertIsInstance(param_b, Sequence)
            for i, callback in enumerate(param_a):
                test_case.assertIsInstance(callback, type(param_b[i]))
        else:
            test_case.assertEqual(param_a, param_b, f"Test failed for {param_name}")



"""Functions to handle sparse matrices."""

from __future__ import annotations

from typing import Iterable

from scipy import sparse


def sparse_from_index_value_dicts(
    row_index_lists: Iterable[dict[int, int]], n_columns: int
) -> sparse.csr_matrix:
    """Create a sparse matrix from list of dicts.

    Each dict represents one row.
    Keys in dictionary correspond to colum index, values represent values of column.

    Parameters
    ----------
    row_index_lists: Iterable[dict[int, int]]
        Iterable of dicts of which each holds column positions and values.
    n_columns: int
        Total number of columns

    Returns
    -------
    sparse.csr_matrix
        Has shape (len(row_index_lists), n_columns).
    """
    data: list[int] = []
    row_positions: list[int] = []
    col_positions: list[int] = []
    row_idx = -1
    for row_idx, row_dict in enumerate(row_index_lists):
        data.extend(row_dict.values())
        col_positions.extend(row_dict.keys())
        row_positions.extend([row_idx] * len(row_dict))
    if row_idx == -1:
        return sparse.csr_matrix((0, n_columns))

    return sparse.csr_matrix(
        (data, (row_positions, col_positions)), shape=(row_idx + 1, n_columns)
    )


def are_equal(matrix_a: sparse.csr_matrix, matrix_b: sparse.csr_matrix) -> bool:
    """Compare if any element is not equal, as this is more efficient.

    Parameters
    ----------
    matrix_a: sparse.csr_matrix
        Matrix A to compare.
    matrix_b: sparse.csr_matrix
        Matrix B to compare.

    Returns
    -------
    bool
        Whether the matrices are equal or not.
    """
    is_unequal_matrix = matrix_a != matrix_b
    number_unequal_elements = int(is_unequal_matrix.nnz)
    return number_unequal_elements == 0



"""Mock PipelineElement for testing."""

from __future__ import annotations

import copy
from typing import Any, Iterable, Optional

import numpy as np

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    TransformingPipelineElement,
)


class MockTransformingPipelineElement(TransformingPipelineElement):
    """Mock element for testing."""

    def __init__(
        self,
        *,
        invalid_values: set[Any] | None = None,
        return_as_numpy_array: bool = False,
        name: str = "dummy",
        uuid: Optional[str] = None,
        n_jobs: int = 1,
    ) -> None:
        """Initialize MockTransformingPipelineElement.

        Parameters
        ----------
        invalid_values: set[Any] | None, optional (default=None)
            Set of values to consider invalid.
        return_as_numpy_array: bool, optional (default=False)
            If True return output as numpy array, otherwise as list.
        name: str, optional (default="dummy")
            Name of PipelineElement
        uuid: str, optional (default=None)
            Unique identifier of PipelineElement.
        n_jobs: int, optional (default=1)
            Number of jobs to run in parallel.
        """
        super().__init__(name=name, uuid=uuid, n_jobs=n_jobs)
        if invalid_values is None:
            invalid_values = set()
        self.invalid_values = invalid_values
        self.return_as_numpy_array: bool = return_as_numpy_array

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all parameters defining the object.
        """
        params = super().get_params(deep)
        if deep:
            params["invalid_values"] = copy.deepcopy(self.invalid_values)
            params["return_as_numpy_array"] = copy.deepcopy(self.return_as_numpy_array)
        else:
            params["invalid_values"] = self.invalid_values
            params["return_as_numpy_array"] = self.return_as_numpy_array
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of the object.

        Parameters
        ----------
        parameters: Any
            Dictionary containing all parameters defining the object.

        Returns
        -------
        Self
            MockTransformingPipelineElement with updated parameters.
        """
        super().set_params(**parameters)
        if "invalid_values" in parameters:
            self.invalid_values = set(parameters["invalid_values"])
        if "return_as_numpy_array" in parameters:
            self.return_as_numpy_array = bool(parameters["return_as_numpy_array"])
        return self

    def pretransform_single(self, value: Any) -> Any:
        """Transform input value to other value.

        Parameters
        ----------
        value: Any
            Input value.

        Returns
        -------
        Any
            Other value.
        """
        if value in self.invalid_values:
            return InvalidInstance(
                self.uuid,
                f"Invalid input value by mock: {value}",
                self.name,
            )
        return value

    def assemble_output(self, value_list: Iterable[Any]) -> Any:
        """Aggregate rows, which in most cases is just return the list.

        Some representations might be better representd as a single object. For example a list of vectors can
        be transformed to a matrix.

        Parameters
        ----------
        value_list: Iterable[Any]
            Iterable of transformed rows.

        Returns
        -------
        Any
            Aggregated output. This can also be the original input.
        """
        if self.return_as_numpy_array:
            return np.array(list(value_list))
        return list(value_list)



"""Initialize Chemprop module."""

try:
    from molpipeline.estimators.chemprop.models import (  # noqa: F401
        ChempropClassifier,
        ChempropModel,
        ChempropNeuralFP,
        ChempropRegressor,
    )

    __all__ = [
        "ChempropClassifier",
        "ChempropModel",
        "ChempropNeuralFP",
        "ChempropRegressor",
    ]
except ImportError:
    __all__ = []



"""Test elements for standardizing molecules."""

import unittest

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2mol import (
    ExplicitHydrogenRemover,
    FragmentDeduplicator,
    IsotopeRemover,
    LargestFragmentChooser,
    MetalDisconnector,
    SaltRemover,
    SolventRemover,
    StereoRemover,
    TautomerCanonicalizer,
    Uncharger,
)

STEREO_MOL_LIST = ["Br[C@@H](Cl)F", "Br[C@H](Cl)F.I[C@@H](Cl)F"]
NON_STEREO_MOL_LIST = ["FC(Cl)Br", "FC(Cl)Br.FC(Cl)I"]


class MolStandardizationTest(unittest.TestCase):
    """Test if the mol2mol standardization elements work as expected."""

    def test_stereo_removal(self) -> None:
        """Test if stereo-information is removed correctly.

        Returns
        -------
        None
        """
        smi2mol = SmilesToMol()
        stereo_removal = StereoRemover()
        mol2smi = MolToSmiles()
        stereo_removal_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("stereo_removal", stereo_removal),
                ("mol2smi", mol2smi),
            ]
        )
        stereo_removed_mol_list = stereo_removal_pipeline.fit_transform(STEREO_MOL_LIST)
        self.assertEqual(stereo_removed_mol_list, NON_STEREO_MOL_LIST)

    def test_metal_disconnector_does_not_lose_ringinfo(self) -> None:
        """Test metal disconnector returns valid molecules containing ring info.

        Returns
        -------
        None
        """

        # example where metal disconnection leads to inconsistent ringinfo -> Sanitization is necessary.
        smiles_uninitialized_ringinfo_after_disconnect = (
            "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"
        )
        smi2mol = SmilesToMol()
        disconnect_metals = MetalDisconnector()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("disconnect_metals", disconnect_metals),
            ]
        )
        mols_processed = pipeline.fit_transform(
            [smiles_uninitialized_ringinfo_after_disconnect]
        )
        self.assertEqual(len(mols_processed), 1)
        # Without additional sanitiziting after disconnecting metals the following would fail with
        # a pre-condition assert from within RDkit.
        self.assertEqual(mols_processed[0].GetRingInfo().NumRings(), 1)

    def test_duplicate_fragment_by_hash_removal(self) -> None:
        """Test if duplicate fragements are correctly removed by DeduplicateFragmentsByInchiElement.

        Returns
        -------
        None
        """
        duplicate_fragment_smiles_lust = [
            "CC.CC.C",
            "c1ccccc1.C1=C-C=C-C=C1",
            "CCC.CC.CC.C",
            "CC1=C[NH]C=N1.CC1=CN=C[NH]1",
        ]
        expected_unique_fragment_smiles_list = [
            "C.CC",
            "c1ccccc1",
            "C.CC.CCC",
            "Cc1c[nH]cn1",
        ]

        smi2mol = SmilesToMol()
        unique_fragments = FragmentDeduplicator()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("unique_fragments", unique_fragments),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(duplicate_fragment_smiles_lust)
        self.assertEqual(expected_unique_fragment_smiles_list, mols_processed)

    def test_tautomer_canonicalization(self) -> None:
        """Test if correct tautomers are generated.

        Returns
        -------
        None
        """
        non_canonical_tautomer_list = [
            "Oc1c(cccc3)c3nc2ccncc12",
            "CN=c1nc[nH]cc1",
            "CN=c1nc[nH]cc1.Oc1c(cccc3)c3nc2ccncc12",
        ]
        canonical_tautomer_list = [
            "O=c1c2ccccc2[nH]c2ccncc12",
            "CNc1ccncn1",
            "CNc1ccncn1.O=c1c2ccccc2[nH]c2ccncc12",
        ]

        smi2mol = SmilesToMol()
        canonical_tautomer = TautomerCanonicalizer()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("canonical_tautomer", canonical_tautomer),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(non_canonical_tautomer_list)
        self.assertEqual(canonical_tautomer_list, mols_processed)

    def test_charge_neutralization(self) -> None:
        """Test if charge neutralization works as expected.

        Returns
        -------
        None
        """
        mol_list = ["CC(=O)-[O-]", "CC(=O)-[O-].C[NH+](C)C"]
        expected_charge_neutralized_smiles_list = ["CC(=O)O", "CC(=O)O.CN(C)C"]

        smi2mol = SmilesToMol()
        charge_neutralizer = Uncharger()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("charge_neutralizer", charge_neutralizer),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_charge_neutralized_smiles_list, mols_processed)

    def test_largest_fragment_chooser_element(self) -> None:
        """Test if largest fragment chooser element works as expected.

        Returns
        -------
        None
        """
        mol_list = ["CCC.CC.C", "c1ccccc1.[Na+].[Cl-]"]
        expected_largest_fragment_smiles_list = ["CCC", "c1ccccc1"]

        smi2mol = SmilesToMol()
        largest_fragment_chooser = LargestFragmentChooser()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("largest_fragment_chooser", largest_fragment_chooser),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_largest_fragment_smiles_list, mols_processed)

    def test_salt_removal_pipeline_element(self) -> None:
        """Test if salt removal pipeline element works as expected.

        Returns
        -------
        None
        """

        mol_list = [
            "[Na+].ClCC(=O)[O-]",
            "[Ca+2].BrCC(=O)[O-].CCC(=O)[O-]",
        ]
        expected_smiles_list = [
            "O=C([O-])CCl",
            "CCC(=O)[O-].O=C([O-])CBr",
        ]
        smitomol = SmilesToMol()
        salt_remover = SaltRemover()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smitomol", smitomol),
                ("salt_remover", salt_remover),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_smiles_list, mols_processed)

    def test_solvent_removal_pipeline_element(self) -> None:
        """Test if solvent removal pipeline element works as expected.

        Returns
        -------
        None
        """

        mol_list = [
            "[OH2].c1ccccc1",
            "ClCCl.c1ccccc1",
            "ClC(Cl)Cl.c1ccccc1",
            "CCOC(=O)C.c1ccccc1",
            "CO.c1ccccc1",
            "CC(C)O.c1ccccc1",
            "CC(=O)C.c1ccccc1",
            "CS(=O)C.c1ccccc1",
            "CCO.c1ccccc1",
            "[OH2].ClCCl.ClC(Cl)Cl.CCOC(=O)C.CO.CC(C)O.CC(=O)C.CS(=O)C.CCO.c1ccccc1",
        ]
        expected_largest_fragment_smiles_list = [
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
            "c1ccccc1",
        ]

        smi2mol = SmilesToMol()
        fragment_remover = SolventRemover()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("fragment_remover", fragment_remover),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_largest_fragment_smiles_list, mols_processed)

    def test_isotopeinfo_removal_pipeline_element(self) -> None:
        """Test if isotopinfo removal pipeline element works as expected.

        Returns
        -------
        None
        """

        mol_list = ["[2H]c1ccccc1", "CC[13CH2][19F]"]
        expected_largest_fragment_smiles_list = ["[H]c1ccccc1", "CCCF"]

        smi2mol = SmilesToMol()
        fragment_remover = IsotopeRemover()
        mol2smi = MolToSmiles()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("isotope_info_remover", fragment_remover),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_largest_fragment_smiles_list, mols_processed)

    def test_explicit_hydrogen_removal_pipeline_element(self) -> None:
        """Test if explicit hydrogen removal pipeline element works as expected.

        Returns
        -------
        None
        """

        mol_list = ["[H]c1ccccc1", "Cc1cncn(-[H])1", "[H][H]"]
        expected_largest_fragment_smiles_list = ["c1ccccc1", "Cc1cnc[nH]1", "[H][H]"]

        smi2mol = SmilesToMol()
        fragment_remover = ExplicitHydrogenRemover()
        mol2smi = MolToSmiles()

        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("explicit_hydrogen_remover", fragment_remover),
                ("mol2smi", mol2smi),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_largest_fragment_smiles_list, mols_processed)


if __name__ == "__main__":
    unittest.main()



"""Variables that are used in multiple tests."""

from chemprop.nn import BCELoss
from torch import Tensor, nn

from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BinaryClassificationFFN,
    BondMessagePassing,
    MulticlassClassificationFFN,
    SumAggregation,
)
from molpipeline.estimators.chemprop.loss_wrapper import CrossEntropyLoss

# These are model parameters which are copied by value, but are too complex to check for equality.
# Thus, for these model parameters, only the type is checked.
NO_IDENTITY_CHECK = [
    "model__agg",
    "model__message_passing",
    "model",
    "model__predictor",
    "model__predictor__criterion",
    "model__predictor__output_transform",
]

# Default parameters for the Chemprop model.

DEFAULT_BINARY_CLASSIFICATION_PARAMS = {
    "batch_size": 64,
    "lightning_trainer": None,
    "lightning_trainer__enable_checkpointing": False,
    "lightning_trainer__enable_model_summary": False,
    "lightning_trainer__max_epochs": 500,
    "lightning_trainer__accelerator": "cpu",
    "lightning_trainer__default_root_dir": None,
    "lightning_trainer__limit_predict_batches": 1.0,
    "lightning_trainer__detect_anomaly": False,
    "lightning_trainer__reload_dataloaders_every_n_epochs": 0,
    "lightning_trainer__precision": "32-true",
    "lightning_trainer__min_steps": None,
    "lightning_trainer__max_time": None,
    "lightning_trainer__limit_train_batches": 1.0,
    "lightning_trainer__strategy": "auto",
    "lightning_trainer__gradient_clip_algorithm": None,
    "lightning_trainer__log_every_n_steps": 50,
    "lightning_trainer__limit_val_batches": 1.0,
    "lightning_trainer__gradient_clip_val": None,
    "lightning_trainer__overfit_batches": 0.0,
    "lightning_trainer__num_nodes": 1,
    "lightning_trainer__use_distributed_sampler": True,
    "lightning_trainer__check_val_every_n_epoch": 1,
    "lightning_trainer__benchmark": False,
    "lightning_trainer__inference_mode": True,
    "lightning_trainer__limit_test_batches": 1.0,
    "lightning_trainer__fast_dev_run": False,
    "lightning_trainer__logger": None,
    "lightning_trainer__max_steps": -1,
    "lightning_trainer__num_sanity_val_steps": 2,
    "lightning_trainer__devices": "auto",
    "lightning_trainer__min_epochs": None,
    "lightning_trainer__val_check_interval": 1.0,
    "lightning_trainer__barebones": False,
    "lightning_trainer__accumulate_grad_batches": 1,
    "lightning_trainer__deterministic": False,
    "lightning_trainer__enable_progress_bar": True,
    "model": MPNN,
    "model__agg__dim": 0,
    "model__agg": SumAggregation,
    "model__batch_norm": True,
    "model__final_lr": 0.0001,
    "model__init_lr": 0.0001,
    "model__max_lr": 0.001,
    "model__message_passing__activation": "relu",
    "model__message_passing__bias": False,
    "model__message_passing__d_e": 14,
    "model__message_passing__d_h": 300,
    "model__message_passing__d_v": 72,
    "model__message_passing__d_vd": None,
    "model__message_passing__depth": 3,
    "model__message_passing__dropout_rate": 0.0,
    "model__message_passing__undirected": False,
    "model__message_passing": BondMessagePassing,
    "model__metric_list": None,
    "model__predictor__activation": "relu",
    "model__warmup_epochs": 2,
    "model__predictor": BinaryClassificationFFN,
    "model__predictor__criterion": BCELoss,
    "model__predictor__criterion__task_weights": Tensor([1.0]),
    "model__predictor__dropout": 0,
    "model__predictor__hidden_dim": 300,
    "model__predictor__input_dim": 300,
    "model__predictor__n_layers": 1,
    "model__predictor__n_tasks": 1,
    "model__predictor__output_transform": nn.Identity,
    "model__predictor__task_weights": Tensor([1.0]),
    "model__predictor__threshold": None,
    "n_jobs": 1,
}

DEFAULT_SET_PARAMS = {
    "batch_size": 64,
    "lightning_trainer": None,
    "lightning_trainer__enable_checkpointing": False,
    "lightning_trainer__enable_model_summary": False,
    "lightning_trainer__max_epochs": 500,
    "lightning_trainer__accelerator": "cpu",
    "lightning_trainer__default_root_dir": None,
    "lightning_trainer__limit_predict_batches": 1.0,
    "lightning_trainer__detect_anomaly": False,
    "lightning_trainer__reload_dataloaders_every_n_epochs": 0,
    "lightning_trainer__precision": "32-true",
    "lightning_trainer__min_steps": None,
    "lightning_trainer__max_time": None,
    "lightning_trainer__limit_train_batches": 1.0,
    "lightning_trainer__strategy": "auto",
    "lightning_trainer__gradient_clip_algorithm": None,
    "lightning_trainer__log_every_n_steps": 50,
    "lightning_trainer__limit_val_batches": 1.0,
    "lightning_trainer__gradient_clip_val": None,
    "lightning_trainer__overfit_batches": 0.0,
    "lightning_trainer__num_nodes": 1,
    "lightning_trainer__use_distributed_sampler": True,
    "lightning_trainer__check_val_every_n_epoch": 1,
    "lightning_trainer__benchmark": False,
    "lightning_trainer__inference_mode": True,
    "lightning_trainer__limit_test_batches": 1.0,
    "lightning_trainer__fast_dev_run": False,
    "lightning_trainer__logger": None,
    "lightning_trainer__max_steps": -1,
    "lightning_trainer__num_sanity_val_steps": 2,
    "lightning_trainer__devices": "auto",
    "lightning_trainer__min_epochs": None,
    "lightning_trainer__val_check_interval": 1.0,
    "lightning_trainer__barebones": False,
    "lightning_trainer__accumulate_grad_batches": 1,
    "lightning_trainer__deterministic": False,
    "lightning_trainer__enable_progress_bar": True,
    "model__agg__dim": 0,
    "model__batch_norm": True,
    "model__final_lr": 0.0001,
    "model__init_lr": 0.0001,
    "model__max_lr": 0.001,
    "model__message_passing__activation": "relu",
    "model__message_passing__bias": False,
    "model__message_passing__d_e": 14,
    "model__message_passing__d_h": 300,
    "model__message_passing__d_v": 72,
    "model__message_passing__d_vd": None,
    "model__message_passing__depth": 3,
    "model__message_passing__dropout_rate": 0.0,
    "model__message_passing__undirected": False,
    "model__metric_list": None,
    "model__predictor__activation": "relu",
    "model__warmup_epochs": 2,
    "model__predictor__dropout": 0,
    "model__predictor__hidden_dim": 300,
    "model__predictor__input_dim": 300,
    "model__predictor__n_layers": 1,
    "model__predictor__n_tasks": 1,
    "model__predictor__threshold": None,
    "n_jobs": 1,
}


DEFAULT_MULTICLASS_CLASSIFICATION_PARAMS = {
    "batch_size": 64,
    "lightning_trainer": None,
    "lightning_trainer__enable_checkpointing": False,
    "lightning_trainer__enable_model_summary": False,
    "lightning_trainer__max_epochs": 500,
    "lightning_trainer__accelerator": "cpu",
    "lightning_trainer__default_root_dir": None,
    "lightning_trainer__limit_predict_batches": 1.0,
    "lightning_trainer__detect_anomaly": False,
    "lightning_trainer__reload_dataloaders_every_n_epochs": 0,
    "lightning_trainer__precision": "32-true",
    "lightning_trainer__min_steps": None,
    "lightning_trainer__max_time": None,
    "lightning_trainer__limit_train_batches": 1.0,
    "lightning_trainer__strategy": "auto",
    "lightning_trainer__gradient_clip_algorithm": None,
    "lightning_trainer__log_every_n_steps": 50,
    "lightning_trainer__limit_val_batches": 1.0,
    "lightning_trainer__gradient_clip_val": None,
    "lightning_trainer__overfit_batches": 0.0,
    "lightning_trainer__num_nodes": 1,
    "lightning_trainer__use_distributed_sampler": True,
    "lightning_trainer__check_val_every_n_epoch": 1,
    "lightning_trainer__benchmark": False,
    "lightning_trainer__inference_mode": True,
    "lightning_trainer__limit_test_batches": 1.0,
    "lightning_trainer__fast_dev_run": False,
    "lightning_trainer__logger": None,
    "lightning_trainer__max_steps": -1,
    "lightning_trainer__num_sanity_val_steps": 2,
    "lightning_trainer__devices": "auto",
    "lightning_trainer__min_epochs": None,
    "lightning_trainer__val_check_interval": 1.0,
    "lightning_trainer__barebones": False,
    "lightning_trainer__accumulate_grad_batches": 1,
    "lightning_trainer__deterministic": False,
    "lightning_trainer__enable_progress_bar": True,
    "model": MPNN,
    "model__agg__dim": 0,
    "model__agg": SumAggregation,
    "model__batch_norm": True,
    "model__final_lr": 0.0001,
    "model__init_lr": 0.0001,
    "model__max_lr": 0.001,
    "model__message_passing__activation": "relu",
    "model__message_passing__bias": False,
    "model__message_passing__d_e": 14,
    "model__message_passing__d_h": 300,
    "model__message_passing__d_v": 72,
    "model__message_passing__d_vd": None,
    "model__message_passing__depth": 3,
    "model__message_passing__dropout_rate": 0.0,
    "model__message_passing__undirected": False,
    "model__message_passing": BondMessagePassing,
    "model__metric_list": None,
    "model__predictor__activation": "relu",
    "model__warmup_epochs": 2,
    "model__predictor": MulticlassClassificationFFN,
    "model__predictor__criterion": CrossEntropyLoss,
    "model__predictor__criterion__task_weights": Tensor([1.0, 1.0, 1.0]),
    "model__predictor__dropout": 0,
    "model__predictor__hidden_dim": 300,
    "model__predictor__input_dim": 300,
    "model__predictor__n_classes": 3,
    "model__predictor__n_layers": 1,
    "model__predictor__n_tasks": 1,
    "model__predictor__output_transform": nn.Identity,
    "model__predictor__task_weights": Tensor([1.0, 1.0, 1.0]),
    "model__predictor__threshold": None,
    "n_classes": 3,
    "n_jobs": 1,
}



"""Tests for the abstract class ABCChemprop."""

import unittest

from molpipeline.estimators.chemprop.abstract import ABCChemprop


class TestABCChemprop(unittest.TestCase):
    """Test static methods of the Chemprop model."""

    def test_filter_params_callback(self) -> None:
        """Test the filter_params_callback method."""
        dummy_params = {
            "callback_modelckpt__monitor": "val_loss",
            "other__param": "value",
        }
        # pylint: disable=protected-access
        other_params, callback_params = ABCChemprop._filter_params(
            dummy_params, "callback_modelckpt"
        )
        # pylint: enable=protected-access
        self.assertEqual(callback_params, {"monitor": "val_loss"})
        self.assertEqual(other_params, {"other__param": "value"})

    def test_filter_params_trainer(self) -> None:
        """Test the filter_params_trainer method."""
        dummy_params = {
            "lightning_trainer__max_epochs": 50,
            "other__param": "value",
        }
        # pylint: disable=protected-access
        other_params, trainer_params = ABCChemprop._filter_params(
            dummy_params, "lightning_trainer"
        )
        # pylint: enable=protected-access
        self.assertEqual(trainer_params, {"max_epochs": 50})
        self.assertEqual(other_params, {"other__param": "value"})



"""Class for transforming molecules to SMILES representations."""

from __future__ import annotations

import abc

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.core import MolToAnyPipelineElement


class MolToStringPipelineElement(MolToAnyPipelineElement, abc.ABC):
    """Abstract class for PipelineElements which transform molecules to integer vectors."""

    _output_type = "str"

    def transform(self, values: list[Chem.Mol]) -> list[str]:
        """Transform the list of molecules to sparse matrix.

        Parameters
        ----------
        values: list[Chem.Mol]
            List of RDKit molecules which are transformed to a string representation.

        Returns
        -------
        list[str]
            List of string representations of the molecules.
        """
        string_list: list[str] = super().transform(values)
        return string_list

    @abc.abstractmethod
    def pretransform_single(self, value: Chem.Mol) -> str:
        """Transform mol to a string.

        Parameters
        ----------
        value: Chem.Mol
            Molecule to be transformed to SMILES representation.
        Returns
        -------
        str
            SMILES representation of molecule.
        """



"""Implementations for the RDKit Path Fingerprint."""

from __future__ import annotations  # for all the python 3.8 users out there.

from typing import Any, Literal, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import copy

from rdkit.Chem import rdFingerprintGenerator

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    MolToRDKitGenFPElement,
)


class Mol2PathFP(
    MolToRDKitGenFPElement
):  # pylint: disable=too-many-instance-attributes
    """Folded Path Fingerprint.

    Feature-mapping to vector-positions is arbitrary.

    """

    # pylint: disable=too-many-arguments,too-many-locals,too-many-positional-arguments
    def __init__(
        self,
        min_path: int = 1,
        max_path: int = 7,
        use_hs: bool = True,
        branched_paths: bool = True,
        use_bond_order: bool = True,
        count_simulation: bool = False,
        count_bounds: Any = None,
        n_bits: int = 2048,
        num_bits_per_feature: int = 2,
        atom_invariants_generator: Any = None,
        counted: bool = False,
        return_as: Literal["sparse", "dense", "explicit_bit_vect"] = "sparse",
        name: str = "Mol2PathFP",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize Mol2PathFP.

        Parameters
        ----------
        min_path: int, optional (default=1)
            Minimum path length.
        max_path: int, optional (default=7)
            Maximum path length.
        use_hs: bool, optional (default=True)
            Include hydrogens (If explicit hydrogens are present in the molecule).
        branched_paths: bool, optional (default=True)
            Include branched paths.
        use_bond_order: bool, optional (default=True)
            Include bond order in path.
        count_simulation: bool, optional (default=False)
            Count simulation.
        count_bounds: Any, optional (default=None)
            Set the bins for the bond count.
        n_bits: int, optional (default=2048)
            Size of fingerprint.
        num_bits_per_feature: int, optional (default=2)
            Number of bits per feature.
        atom_invariants_generator: Any, optional (default=None)
            Atom invariants generator.
        counted: bool, optional (default=False)
            If True, the fingerprint will be counted.
            If False, the fingerprint will be binary.
        return_as: Literal["sparse", "dense", "explicit_bit_vect"]
            Type of output. When "sparse" the fingerprints will be returned as a scipy.sparse.csr_matrix
            holding a sparse representation of the bit vectors. With "dense" a numpy matrix will be returned.
            With "explicit_bit_vect" the fingerprints will be returned as a list of RDKit's
            rdkit.DataStructs.cDataStructs.ExplicitBitVect.
        name: str, optional (default="MolToMorganFP")
            Name of PipelineElement
        n_jobs: int, optional (default=1)
            Number of cores to use.
        uuid: str | None, optional (default=None)
            UUID of the PipelineElement.

        References
        ----------
        [1] https://www.rdkit.org/docs/source/rdkit.Chem.rdFingerprintGenerator.html#rdkit.Chem.rdFingerprintGenerator.GetRDKitFPGenerator
        """
        # pylint: disable=R0801
        super().__init__(
            counted=counted,
            return_as=return_as,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        if not isinstance(n_bits, int) or n_bits < 1:
            raise ValueError(
                f"Number of bits has to be a positive integer, which is > 0! (Received: {n_bits})"
            )
        self._n_bits = n_bits
        self._feature_names = [f"path_{i}" for i in range(self._n_bits)]
        self._min_path = min_path
        self._max_path = max_path
        self._use_hs = use_hs
        self._branched_paths = branched_paths
        self._use_bond_order = use_bond_order
        self._count_simulation = count_simulation
        self._count_bounds = count_bounds
        self._num_bits_per_feature = num_bits_per_feature
        self._atom_invariants_generator = atom_invariants_generator

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameters.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["min_path"] = int(self._min_path)
            parameters["max_path"] = int(self._max_path)
            parameters["use_hs"] = bool(self._use_hs)
            parameters["branched_paths"] = bool(self._branched_paths)
            parameters["use_bond_order"] = bool(self._use_bond_order)
            parameters["count_simulation"] = bool(self._count_simulation)
            parameters["count_bounds"] = copy.copy(self._count_bounds)
            parameters["num_bits_per_feature"] = int(self._num_bits_per_feature)
            parameters["atom_invariants_generator"] = copy.copy(
                self._atom_invariants_generator
            )
            parameters["n_bits"] = int(self._n_bits)
        else:
            parameters["min_path"] = self._min_path
            parameters["max_path"] = self._max_path
            parameters["use_hs"] = self._use_hs
            parameters["branched_paths"] = self._branched_paths
            parameters["use_bond_order"] = self._use_bond_order
            parameters["count_simulation"] = self._count_simulation
            parameters["count_bounds"] = self._count_bounds
            parameters["num_bits_per_feature"] = self._num_bits_per_feature
            parameters["atom_invariants_generator"] = self._atom_invariants_generator
            parameters["n_bits"] = self._n_bits
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            MolToMorganFP pipeline element with updated parameters.
        """
        parameter_copy = dict(parameters)
        min_path = parameter_copy.pop("min_path", None)
        if min_path is not None:
            self._min_path = min_path
        max_path = parameter_copy.pop("max_path", None)
        if max_path is not None:
            self._max_path = max_path
        use_hs = parameter_copy.pop("use_hs", None)
        if use_hs is not None:
            self._use_hs = use_hs
        branched_paths = parameter_copy.pop("branched_paths", None)
        if branched_paths is not None:
            self._branched_paths = branched_paths
        use_bond_order = parameter_copy.pop("use_bond_order", None)
        if use_bond_order is not None:
            self._use_bond_order = use_bond_order
        count_simulation = parameter_copy.pop("count_simulation", None)
        if count_simulation is not None:
            self._count_simulation = count_simulation
        count_bounds = parameter_copy.pop("count_bounds", None)
        if count_bounds is not None:
            self._count_bounds = count_bounds
        num_bits_per_feature = parameter_copy.pop("num_bits_per_feature", None)
        if num_bits_per_feature is not None:
            self._num_bits_per_feature = num_bits_per_feature
        atom_invariants_generator = parameter_copy.pop(
            "atom_invariants_generator", None
        )
        if atom_invariants_generator is not None:
            self._atom_invariants_generator = atom_invariants_generator
        n_bits = parameter_copy.pop("n_bits", None)  # pylint: disable=duplicate-code
        if n_bits is not None:
            self._n_bits = n_bits
        super().set_params(**parameter_copy)
        return self

    def _get_fp_generator(self) -> rdFingerprintGenerator.FingerprintGenerator64:
        """Get the fingerprint generator for the RDKit path fingerprint.

        Returns
        -------
        rdFingerprintGenerator.GetRDKitFPGenerator
            RDKit Path fingerprint generator.
        """
        return rdFingerprintGenerator.GetRDKitFPGenerator(
            minPath=self._min_path,
            maxPath=self._max_path,
            fpSize=self._n_bits,
            useHs=self._use_hs,
            branchedPaths=self._branched_paths,
            useBondOrder=self._use_bond_order,
            countSimulation=self._count_simulation,
            countBounds=self._count_bounds,
            numBitsPerFeature=self._num_bits_per_feature,
            atomInvariantsGenerator=self._atom_invariants_generator,
        )



"""Module for explanation class."""

from __future__ import annotations

import abc
import dataclasses

import numpy as np
import numpy.typing as npt

from molpipeline.abstract_pipeline_elements.core import RDKitMol


@dataclasses.dataclass(kw_only=True)
class _AbstractMoleculeExplanation(abc.ABC):
    """Abstract class representing an explanation for a prediction for a molecule."""

    molecule: RDKitMol | None = None
    prediction: npt.NDArray[np.float64] | None = None


@dataclasses.dataclass(kw_only=True)
class FeatureInfoMixin:
    """Mixin providing additional information about the features used in the explanation."""

    feature_vector: npt.NDArray[np.float64] | None = None
    feature_names: list[str] | None = None


@dataclasses.dataclass(kw_only=True)
class FeatureExplanationMixin:
    """Explanation based on feature importance scores, e.g. Shapley Values."""

    # explanation scores for individual features
    feature_weights: npt.NDArray[np.float64] | None = None


@dataclasses.dataclass(kw_only=True)
class AtomExplanationMixin:
    """Atom score based explanation."""

    # explanation scores for individual atoms
    atom_weights: npt.NDArray[np.float64] | None = None


@dataclasses.dataclass(kw_only=True)
class BondExplanationMixin:
    """Bond score based explanation."""

    # explanation scores for individual bonds
    bond_weights: npt.NDArray[np.float64] | None = None


@dataclasses.dataclass(kw_only=True)
class SHAPExplanationMixin:
    """Mixin providing additional information only present in SHAP explanations."""

    expected_value: npt.NDArray[np.float64] | None = None


@dataclasses.dataclass(kw_only=True)
class SHAPFeatureExplanation(
    FeatureInfoMixin,
    FeatureExplanationMixin,
    SHAPExplanationMixin,
    _AbstractMoleculeExplanation,  # base-class should be the last element https://www.ianlewis.org/en/mixins-and-python
):
    """Explanation using feature importance scores from SHAP."""

    def is_valid(self) -> bool:
        """Check if the explanation is valid.

        Returns
        -------
        bool
            True if the explanation is valid, False otherwise.
        """
        return all(
            [
                self.feature_vector is not None,
                self.feature_names is not None,
                self.molecule is not None,
                self.prediction is not None,
                self.feature_weights is not None,
            ]
        )


@dataclasses.dataclass(kw_only=True)
class SHAPFeatureAndAtomExplanation(
    FeatureInfoMixin,
    FeatureExplanationMixin,
    SHAPExplanationMixin,
    AtomExplanationMixin,
    _AbstractMoleculeExplanation,
):
    """Explanation using feature and atom importance scores from SHAP."""

    def is_valid(self) -> bool:
        """Check if the explanation is valid.

        Returns
        -------
        bool
            True if the explanation is valid, False otherwise.
        """
        return all(
            [
                self.feature_vector is not None,
                self.feature_names is not None,
                self.molecule is not None,
                self.prediction is not None,
                self.feature_weights is not None,
                self.atom_weights is not None,
            ]
        )



"""Test functionality set at package init."""

import unittest

from molpipeline import __version__


class TestInit(unittest.TestCase):
    """Test functionality set at package init."""

    def test_version(self) -> None:
        """Test that the package has a version."""
        self.assertIsInstance(__version__, str)
        splitted = __version__.split(".")
        self.assertEqual(len(splitted), 3)
        major, minor, patch = splitted
        self.assertTrue(major.isdigit())
        self.assertTrue(minor.isdigit())
        self.assertTrue(patch.isdigit())



"""Functions for comparing pipelines."""

from typing import Any, TypeVar

from molpipeline import Pipeline
from molpipeline.utils.json_operations import recursive_to_json

_T = TypeVar("_T", list[Any], tuple[Any, ...], set[Any], dict[Any, Any], Any)


def remove_irrelevant_params(params: _T) -> _T:
    """Remove irrelevant parameters from a dictionary.

    Parameters
    ----------
    params : TypeVar
        Parameters to remove irrelevant parameters from.

    Returns
    -------
    TypeVar
        Parameters without irrelevant parameters.
    """
    if isinstance(params, list):
        return [remove_irrelevant_params(val) for val in params]
    if isinstance(params, tuple):
        return tuple(remove_irrelevant_params(val) for val in params)
    if isinstance(params, set):
        return {remove_irrelevant_params(val) for val in params}

    irrelevant_params = ["n_jobs", "uuid", "error_filter_id"]
    if isinstance(params, dict):
        params_new = {}
        for key, value in params.items():
            if not isinstance(key, str):
                continue
            if key.split("__")[-1] in irrelevant_params:
                continue
            params_new[key] = remove_irrelevant_params(value)
        return params_new
    return params


def compare_recursive(  # pylint: disable=too-many-return-statements
    value_a: Any, value_b: Any
) -> bool:
    """Compare two values recursively.

    Parameters
    ----------
    value_a : Any
        First value to compare.
    value_b : Any
        Second value to compare.

    Returns
    -------
    bool
        True if the values are the same, False otherwise.
    """
    if value_a.__class__ != value_b.__class__:
        return False

    if isinstance(value_a, dict):
        if set(value_a.keys()) != set(value_b.keys()):
            return False
        for key in value_a:
            if not compare_recursive(value_a[key], value_b[key]):
                return False
        return True

    if isinstance(value_a, (list, tuple)):
        if len(value_a) != len(value_b):
            return False
        for val_a, val_b in zip(value_a, value_b):
            if not compare_recursive(val_a, val_b):
                return False
        return True
    return value_a == value_b


def check_pipelines_equivalent(pipeline_a: Pipeline, pipeline_b: Pipeline) -> bool:
    """Check if two pipelines are the same.

    Parameters
    ----------
    pipeline_a : Pipeline
        Pipeline to compare.
    pipeline_b : Pipeline
        Pipeline to compare.

    Returns
    -------
    bool
        True if the pipelines are the same, False otherwise.
    """
    if not isinstance(pipeline_a, Pipeline) or not isinstance(pipeline_b, Pipeline):
        raise ValueError("Both inputs should be of type Pipeline.")
    pipeline_json_a = recursive_to_json(pipeline_a)
    pipeline_json_a = remove_irrelevant_params(pipeline_json_a)
    pipeline_json_b = recursive_to_json(pipeline_b)
    pipeline_json_b = remove_irrelevant_params(pipeline_json_b)
    return compare_recursive(pipeline_json_a, pipeline_json_b)



"""Init."""

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    MolToFingerprintPipelineElement,
)
from molpipeline.abstract_pipeline_elements.mol2any.mol2floatvector import (
    MolToDescriptorPipelineElement,
)
from molpipeline.abstract_pipeline_elements.mol2any.mol2string import (
    MolToStringPipelineElement,
)

__all__ = (
    "MolToFingerprintPipelineElement",
    "MolToDescriptorPipelineElement",
    "MolToStringPipelineElement",
)



"""Wrap Chemprop in a sklearn like transformer returning the neural fingerprint as a numpy array."""

from typing import Any, Self, Sequence

import numpy as np
import numpy.typing as npt
from chemprop.data import BatchMolGraph, MoleculeDataset
from chemprop.models.model import MPNN
from lightning import pytorch as pl

from molpipeline.estimators.chemprop.abstract import ABCChemprop


class ChempropNeuralFP(ABCChemprop):
    """Wrap Chemprop in a sklearn like transformer returning the neural fingerprint as a numpy array.

    This class is not a (grand-) child of MolToAnyPipelineElement, as it does not support the `pretransform_single`
    method. To maintain compatibility with the MolToAnyPipelineElement, the `output_type` property is implemented.
    It can be used as any other transformer in the pipeline, except in the `MolToConcatenatedVector`.
    """

    @property
    def output_type(self) -> str:
        """Return the output type of the transformer."""
        return "float"

    def __init__(
        self,
        model: MPNN,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        disable_fitting: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop neural fingerprint model.

        Parameters
        ----------
        model : MPNN
            The chemprop model to wrap.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        disable_fitting : bool, optional (default=False)
            Whether to allow fitting or set to fixed encoding.
        **kwargs: Any
            Parameters for components of the model.
        """
        # pylint: disable=duplicate-code
        self.disable_fitting = disable_fitting
        super().__init__(
            model=model,
            lightning_trainer=lightning_trainer,
            batch_size=batch_size,
            n_jobs=n_jobs,
            **kwargs,
        )

    def fit(
        self,
        X: MoleculeDataset,
        y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64],
    ) -> Self:
        """Fit the model.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            The target data.

        Returns
        -------
        Self
            The fitted model.
        """
        if self.disable_fitting:
            return self
        return super().fit(X, y)

    def transform(
        self, X: MoleculeDataset  # pylint: disable=invalid-name
    ) -> npt.NDArray[np.float64]:
        """Transform the input.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.

        Returns
        -------
        npt.NDArray[np.float64]
            The neural fingerprint of the input data.
        """
        self.model.eval()
        mol_data = [X[i].mg for i in range(len(X))]
        return self.model.fingerprint(BatchMolGraph(mol_data)).detach().numpy()

    def fit_transform(
        self,
        X: MoleculeDataset,  # pylint: disable=invalid-name
        y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64],
    ) -> npt.NDArray[np.float64]:
        """Fit the model and transform the input.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            The target data.

        Returns
        -------
        npt.NDArray[np.float64]
            The neural fingerprint of the input data.
        """
        self.fit(X, y)
        return self.transform(X)



"""Test cases for the nearest neighbors estimators."""

from unittest import TestCase

import numpy as np
from sklearn.base import clone

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.any2mol import SmilesToMol
from molpipeline.estimators import NamedNearestNeighbors, TanimotoToTraining
from molpipeline.mol2any import MolToMorganFP
from molpipeline.utils.kernel import tanimoto_distance_sparse

TEST_SMILES = [
    "c1ccccc1",
    "c1cc(-C(=O)O)ccc1",
    "CCCCCCN",
    "CCCCCCO",
]

TWO_NN = [
    ["c1ccccc1", "c1cc(-C(=O)O)ccc1"],
    ["c1cc(-C(=O)O)ccc1", "c1ccccc1"],
    ["CCCCCCN", "CCCCCCO"],
    ["CCCCCCO", "CCCCCCN"],
]

TWO_NN_SIMILARITIES = np.array(
    [
        [1.0, 3 / 14],
        [1.0, 3 / 14],
        [1.0, 4 / 9],
        [1.0, 4 / 9],
    ]
)


class TestNamedNearestNeighbors(TestCase):
    """Test the NamedNearestNeighbors class if correct names are returned."""

    def test_fit_predict_one(self) -> None:
        """Test the fit_predict method with one prediced NN."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=1, metric="jaccard")),
            ]
        )
        result = model.fit_predict(TEST_SMILES, TEST_SMILES).tolist()
        self.assertListEqual(result, [[nn_list[0]] for nn_list in TWO_NN])

    def test_fit_predict_two(self) -> None:
        """Test the fit_predict method with two prediced NN."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="jaccard")),
            ]
        )
        result = model.fit_predict(TEST_SMILES, TEST_SMILES).tolist()
        self.assertListEqual(result, TWO_NN)

    def test_fit_and_predict_with_distance(self) -> None:
        """Test the fit_predict method with distance."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="jaccard")),
            ]
        )
        model.fit(TEST_SMILES, TEST_SMILES)
        result = model.predict(TEST_SMILES, **{"return_distance": True})
        neighbors = result[:, :, 0]
        distances = result[:, :, 1].astype(np.float64)
        self.assertListEqual(neighbors.tolist(), TWO_NN)
        self.assertTrue(np.allclose(1 - distances, TWO_NN_SIMILARITIES))

    def test_fit_predict_with_n_neigbours(self) -> None:
        """Test the fit_predict method with parameter n_neighbors."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=1, metric="jaccard")),
            ]
        )
        result = model.fit_predict(
            TEST_SMILES, TEST_SMILES, lookup__n_neighbors=2
        ).tolist()
        self.assertListEqual(result, TWO_NN)

    def test_fit_and_predict_with_n_neigbours(self) -> None:
        """Test the fit_predict method with parameter n_neighbors."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=1, metric="jaccard")),
            ]
        )
        model.fit(TEST_SMILES, TEST_SMILES)
        result = model.predict(TEST_SMILES, **{"n_neighbors": 2}).tolist()
        self.assertListEqual(result, TWO_NN)

    def test_fit_and_predict_with_fit_predict(self) -> None:
        """Test if the fit_predict method gives same results as calling fit and predict separately."""
        model1 = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="jaccard")),
            ]
        )
        model2 = clone(model1)
        result1 = model1.fit_predict(TEST_SMILES, TEST_SMILES).tolist()
        model2.fit(TEST_SMILES, TEST_SMILES)
        result2 = model2.predict(TEST_SMILES).tolist()
        self.assertListEqual(result1, TWO_NN)
        self.assertListEqual(result2, TWO_NN)
        self.assertListEqual(result1, result2)

    def test_fit_predict_custom_metric(self) -> None:
        """Test if the custom metric can be used."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP(return_as="sparse")),
                (
                    "lookup",
                    NamedNearestNeighbors(
                        n_neighbors=2, metric=tanimoto_distance_sparse
                    ),
                ),
            ]
        )
        result = model.fit_predict(TEST_SMILES, TEST_SMILES).tolist()
        self.assertListEqual(result, TWO_NN)

    def test_fit_predict_precomputed(self) -> None:
        """Test the fit_predict method with precomputed similarities."""
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("fingerprint", MolToMorganFP()),
                ("tanimoto", TanimotoToTraining(distance=True)),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="precomputed")),
            ]
        )
        result = model.fit_predict(TEST_SMILES, TEST_SMILES).tolist()
        self.assertListEqual(result, TWO_NN)

    def test_fit_predict_invalid(self) -> None:
        """Test the fit_predict method with invalid smiles."""

        with_invald_smiles = ["CC1CC"] + TEST_SMILES

        error_filter = ErrorFilter(filter_everything=True)
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="jaccard")),
                (
                    "error_replacer",
                    PostPredictionWrapper(
                        FilterReinserter.from_error_filter(
                            error_filter, fill_value="invalid"
                        )
                    ),
                ),
            ]
        )
        result = model.fit_predict(with_invald_smiles, with_invald_smiles).tolist()
        self.assertListEqual(result, [["invalid", "invalid"]] + TWO_NN)

        result_only_valid = model.predict(TEST_SMILES).tolist()
        self.assertListEqual(result_only_valid, TWO_NN)

    def test_fit_and_predict_invalid_with_distance(self) -> None:
        """Test the fit_predict method with invalid smiles and distance."""
        with_invald_smiles = ["CC1CC"] + TEST_SMILES

        error_filter = ErrorFilter(filter_everything=True)
        model = Pipeline(
            [
                ("mol", SmilesToMol()),
                ("error_filter", error_filter),
                ("fingerprint", MolToMorganFP(return_as="dense")),
                ("lookup", NamedNearestNeighbors(n_neighbors=2, metric="jaccard")),
                (
                    "error_replacer",
                    PostPredictionWrapper(
                        FilterReinserter.from_error_filter(
                            error_filter, fill_value="invalid"
                        )
                    ),
                ),
            ]
        )
        model.fit(with_invald_smiles, with_invald_smiles)
        result = model.predict(with_invald_smiles, **{"return_distance": True})
        neighbors = result[:, :, 0]
        distances = result[:, :, 1]
        self.assertListEqual(neighbors.tolist(), [["invalid", "invalid"]] + TWO_NN)
        self.assertTrue(
            1 - np.allclose(distances[1:, :].astype(np.float64), TWO_NN_SIMILARITIES)
        )



"""Implementations for the Morgan fingerprint."""

from __future__ import annotations  # for all the python 3.8 users out there.

from typing import Any, Literal, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import copy

from rdkit.Chem import AllChem, rdFingerprintGenerator

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    ABCMorganFingerprintPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToMorganFP(ABCMorganFingerprintPipelineElement):
    """Folded Morgan Fingerprint.

    Feature-mapping to vector-positions is arbitrary.

    """

    # pylint: disable=R0913
    def __init__(
        self,
        radius: int = 2,
        use_features: bool = False,
        n_bits: int = 2048,
        counted: bool = False,
        return_as: Literal["sparse", "dense", "explicit_bit_vect"] = "sparse",
        name: str = "MolToMorganFP",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MolToMorganFP.

        Parameters
        ----------
        radius: int, optional (default=2)
            radius of the circular fingerprint [1]. Radius of 2 corresponds to ECFP4 (radius 2 -> diameter 4)
        use_features: bool, optional (default=False)
            Instead of atoms, features are encoded in the fingerprint. [2]
        n_bits: int, optional (default=2048)
            Size of fingerprint.
        counted: bool, optional (default=False)
            If True, the fingerprint will be counted.
            If False, the fingerprint will be binary.
        return_as: Literal["sparse", "dense", "explicit_bit_vect"]
            Type of output. When "sparse" the fingerprints will be returned as a scipy.sparse.csr_matrix
            holding a sparse representation of the bit vectors. With "dense" a numpy matrix will be returned.
            With "explicit_bit_vect" the fingerprints will be returned as a list of RDKit's
            rdkit.DataStructs.cDataStructs.ExplicitBitVect.
        name: str, optional (default="MolToMorganFP")
            Name of PipelineElement
        n_jobs: int, optional (default=1)
            Number of cores to use.
        uuid: str | None, optional (default=None)
            UUID of the PipelineElement.

        References
        ----------
            [1] https://rdkit.org/docs/GettingStartedInPython.html#morgan-fingerprints-circular-fingerprints
            [2] https://rdkit.org/docs/GettingStartedInPython.html#feature-definitions-used-in-the-morgan-fingerprints
        """
        # pylint: disable=R0801
        super().__init__(
            radius=radius,
            use_features=use_features,
            counted=counted,
            return_as=return_as,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        if not isinstance(n_bits, int) or n_bits < 1:
            raise ValueError(
                f"Number of bits has to be a positve integer, which is > 0! (Received: {n_bits})"
            )
        self._n_bits = n_bits
        self._feature_names = [f"morgan_{i}" for i in range(self._n_bits)]

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameters.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["n_bits"] = copy.copy(self._n_bits)
        else:
            parameters["n_bits"] = self._n_bits
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            MolToMorganFP pipeline element with updated parameters.
        """
        parameter_copy = dict(parameters)
        n_bits = parameter_copy.pop("n_bits", None)
        if n_bits is not None:
            self._n_bits = n_bits
        super().set_params(**parameter_copy)

        return self

    def _get_fp_generator(
        self,
    ) -> rdFingerprintGenerator.FingerprintGenerator64:
        """Get the fingerprint generator.

        Returns
        -------
        rdFingerprintGenerator.FingerprintGenerator
            RDKit fingerprint generator.
        """
        return rdFingerprintGenerator.GetMorganGenerator(
            radius=self.radius,
            fpSize=self._n_bits,
        )

    def _explain_rdmol(self, mol_obj: RDKitMol) -> dict[int, list[tuple[int, int]]]:
        """Get central atom and radius of all features in molecule.

        Parameters
        ----------
        mol_obj: RDKitMol
            RDKit molecule object

        Returns
        -------
        dict[int, list[tuple[int, int]]]
            Dictionary with bit position as key and list of tuples with atom index and radius as value.
        """
        fp_generator = self._get_fp_generator()
        additional_output = AllChem.AdditionalOutput()
        additional_output.AllocateBitInfoMap()
        # using the dense fingerprint here, to get indices after folding
        _ = fp_generator.GetFingerprint(mol_obj, additionalOutput=additional_output)
        bit_info = additional_output.GetBitInfoMap()
        return bit_info



"""Utils for explainability tests."""

from typing import Any

import scipy

from molpipeline import Pipeline
from molpipeline.utils.subpipeline import get_featurization_subpipeline


def construct_kernel_shap_kwargs(pipeline: Pipeline, data: list[str]) -> dict[str, Any]:
    """Construct the kwargs for SHAPKernelExplainer.

    Convert sparse matrix to dense array because SHAPKernelExplainer does not
    support sparse matrix as `data` and then explain dense matrices.
    We stick to dense matrices for simplicity.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline used for featurization.
    data : list[str]
        The input data, e.g. SMILES strings.

    Returns
    -------
    dict[str, Any]
        The kwargs for SHAPKernelExplainer
    """
    featurization_subpipeline = get_featurization_subpipeline(
        pipeline, raise_not_found=True
    )
    data_transformed = featurization_subpipeline.transform(data)  # type: ignore[union-attr]
    if scipy.sparse.issparse(data_transformed):
        data_transformed = data_transformed.toarray()
    return {"data": data_transformed}



"""Classes for transforming rdkit molecules to inchi."""

from __future__ import annotations

from typing import Optional

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.mol2any.mol2string import (
    MolToStringPipelineElement as _MolToStringPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToInchi(_MolToStringPipelineElement):
    """PipelineElement to transform a molecule to an INCHI string."""

    def pretransform_single(self, value: RDKitMol) -> str:
        """Transform a molecule to a INCHI-key string.

        Parameters
        ----------
        value: RDKitMol
            molecule to transform

        Returns
        -------
        str
            INCHI string
        """
        return str(Chem.MolToInchi(value))


class MolToInchiKey(_MolToStringPipelineElement):
    """PipelineElement to transform a molecule to an INCHI-Key string."""

    def __init__(
        self,
        name: str = "MolToInchiKey",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ):
        """Initialize MolToInchiKey.

        Parameters
        ----------
        name: str
            name of PipelineElement
        n_jobs: int
            number of jobs to use for parallelization
        uuid: Optional[str], optional
            uuid of PipelineElement, by default None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)

    def pretransform_single(self, value: RDKitMol) -> str:
        """Transform a molecule to an INCHI-key string.

        Parameters
        ----------
        value: RDKitMol
            molecule to transform

        Returns
        -------
        str
            INCHI-key of molecule.
        """
        return str(Chem.MolToInchiKey(value))



"""Tests for the MolToFoldedMorganFingerprint pipeline element."""

from __future__ import annotations

import unittest
from typing import Any

import numpy as np

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToMorganFP
from tests.utils.fingerprints import fingerprints_to_numpy

test_smiles = [
    "c1ccccc1",
    "c1ccccc1C",
    "NCCOCCCC(=O)O",
]


class TestMol2MorganFingerprint(unittest.TestCase):
    """Unittest for MolToFoldedMorganFingerprint, which calculates folded Morgan Fingerprints."""

    def test_can_be_constructed(self) -> None:
        """Test if the MolToFoldedMorganFingerprint pipeline element can be constructed.

        Returns
        -------
        None
        """
        mol_fp = MolToMorganFP()
        mol_fp_copy = mol_fp.copy()
        self.assertTrue(mol_fp_copy is not mol_fp)
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_copy.get_params()[key])
        mol_fp_recreated = MolToMorganFP(**mol_fp.get_params())
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_recreated.get_params()[key])

    def test_counted_bits(self) -> None:
        """Test if the option counted bits works as expected.

        Returns
        -------
        None
        """
        mol_fp = MolToMorganFP(radius=2, n_bits=1024, return_as="dense")
        smi2mol = SmilesToMol()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol_fp", mol_fp),
            ],
        )
        output_binary = pipeline.fit_transform(test_smiles)
        pipeline.set_params(mol_fp__counted=True)
        output_counted = pipeline.fit_transform(test_smiles)
        self.assertTrue(
            np.all(np.flatnonzero(output_counted) == np.flatnonzero(output_binary))
        )
        self.assertTrue(np.all(output_counted >= output_binary))
        self.assertTrue(np.any(output_counted > output_binary))

    def test_output_types(self) -> None:
        """Test if the output types are correct for counted and binary fingerprints."""
        mol_fp = MolToMorganFP(counted=False)
        self.assertEqual(mol_fp.output_type, "binary")
        mol_fp = MolToMorganFP(counted=True)
        self.assertEqual(mol_fp.output_type, "integer")

    def test_return_value_types(self) -> None:
        """Test equality of different output_types."""

        smi2mol = SmilesToMol()
        sparse_morgan = MolToMorganFP(radius=2, n_bits=1024, return_as="sparse")
        dense_morgan = MolToMorganFP(radius=2, n_bits=1024, return_as="dense")
        explicit_bit_vect_morgan = MolToMorganFP(
            radius=2, n_bits=1024, return_as="explicit_bit_vect"
        )
        sparse_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("sparse_morgan", sparse_morgan),
            ],
        )
        dense_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("dense_morgan", dense_morgan),
            ],
        )
        explicit_bit_vect_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("explicit_bit_vect_morgan", explicit_bit_vect_morgan),
            ],
        )

        sparse_output = sparse_pipeline.fit_transform(test_smiles)
        dense_output = dense_pipeline.fit_transform(test_smiles)
        explicit_bit_vect_morgan_output = explicit_bit_vect_pipeline.fit_transform(
            test_smiles
        )

        self.assertTrue(np.all(sparse_output.toarray() == dense_output))

        self.assertTrue(
            np.equal(
                dense_output,
                np.array(explicit_bit_vect_morgan_output),
            ).all()
        )

    def test_setter_getter(self) -> None:
        """Test if the setters and getters work as expected."""
        mol_fp = MolToMorganFP()
        params: dict[str, Any] = {
            "radius": 2,
            "n_bits": 1024,
            "return_as": "dense",
        }
        mol_fp.set_params(**params)
        self.assertEqual(mol_fp.get_params()["radius"], 2)
        self.assertEqual(mol_fp.get_params()["n_bits"], 1024)
        self.assertEqual(mol_fp.get_params()["return_as"], "dense")

    def test_setter_getter_error_handling(self) -> None:
        """Test if the setters and getters work as expected when errors are encountered."""

        mol_fp = MolToMorganFP()
        params: dict[str, Any] = {
            "radius": 2,
            "n_bits": 1024,
            "return_as": "invalid-option__11!",
        }
        self.assertRaises(ValueError, mol_fp.set_params, **params)

    def test_bit2atom_mapping(self) -> None:
        """Test that the mapping from bits to atom weights works as intended."""
        n_bits = 2048
        sparse_morgan = MolToMorganFP(radius=2, n_bits=n_bits, return_as="sparse")
        dense_morgan = MolToMorganFP(radius=2, n_bits=n_bits, return_as="dense")
        explicit_bit_vect_morgan = MolToMorganFP(
            radius=2, n_bits=n_bits, return_as="explicit_bit_vect"
        )

        smi2mol = SmilesToMol()
        for test_smi in test_smiles:
            for fp_gen in [sparse_morgan, dense_morgan, explicit_bit_vect_morgan]:
                for counted in [False, True]:
                    mol = smi2mol.transform([test_smi])[0]
                    fp_gen.set_params(counted=counted)
                    fp = fp_gen.transform([mol])
                    mapping = fp_gen.bit2atom_mapping(mol)
                    np_fp = fingerprints_to_numpy(fp)
                    self.assertEqual(np.nonzero(np_fp)[0].shape[0], len(mapping))  # type: ignore

    def test_feature_names(self) -> None:
        """Test if the feature names are correct."""
        mol_fp = MolToMorganFP(n_bits=1024)
        feature_names = mol_fp.feature_names
        self.assertEqual(len(feature_names), 1024)
        # feature names should be unique
        self.assertEqual(len(feature_names), len(set(feature_names)))


if __name__ == "__main__":
    unittest.main()



"""Test gaussian grid visualization."""

import unittest
from typing import ClassVar

import numpy as np
from rdkit import Chem
from rdkit.Chem import Draw

from molpipeline import Pipeline
from molpipeline.experimental.explainability import (
    SHAPFeatureAndAtomExplanation,
    SHAPFeatureExplanation,
    SHAPTreeExplainer,
)
from molpipeline.experimental.explainability.visualization.visualization import (
    make_sum_of_gaussians_grid,
)
from tests.test_experimental.test_explainability.test_visualization.test_visualization import (
    _get_test_morgan_rf_pipeline,
)

TEST_SMILES = ["CC", "CCO", "COC", "c1ccccc1(N)", "CCC(-O)O", "CCCN"]
CONTAINS_OX = [0, 1, 1, 0, 1, 0]


class TestSumOfGaussiansGrid(unittest.TestCase):
    """Test sum of gaussian grid ."""

    # pylint: disable=duplicate-code
    test_pipeline: ClassVar[Pipeline]
    test_explainer: ClassVar[SHAPTreeExplainer]
    test_explanations: ClassVar[
        list[SHAPFeatureAndAtomExplanation | SHAPFeatureExplanation]
    ]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the tests."""
        cls.test_pipeline = _get_test_morgan_rf_pipeline()
        cls.test_pipeline.fit(TEST_SMILES, CONTAINS_OX)
        cls.test_explainer = SHAPTreeExplainer(cls.test_pipeline)
        cls.test_explanations = cls.test_explainer.explain(TEST_SMILES)

    def test_grid_with_shap_atom_weights(self) -> None:
        """Test grid with SHAP atom weights."""
        for explanation in self.test_explanations:
            self.assertTrue(explanation.is_valid())
            self.assertIsInstance(explanation.atom_weights, np.ndarray)  # type: ignore[union-attr]

            mol_copy = Chem.Mol(explanation.molecule)
            mol_copy = Draw.PrepareMolForDrawing(mol_copy)
            value_grid = make_sum_of_gaussians_grid(
                mol_copy,
                atom_weights=explanation.atom_weights,  # type: ignore[union-attr]
                atom_width=np.inf,
                grid_resolution=[8, 8],
                padding=[0.4, 0.4],
            )
            self.assertIsNotNone(value_grid)
            self.assertEqual(value_grid.values.size, 8 * 8)

            # test that the range of summed gaussian values is as expected for SHAP
            self.assertTrue(value_grid.values.min() >= -1)
            self.assertTrue(value_grid.values.max() <= 1)



"""Init the module for mol2mol pipeline elements."""

from molpipeline.mol2mol.filter import (
    ComplexFilter,
    ElementFilter,
    EmptyMoleculeFilter,
    InorganicsFilter,
    MixtureFilter,
    RDKitDescriptorsFilter,
    SmartsFilter,
    SmilesFilter,
)
from molpipeline.mol2mol.reaction import MolToMolReaction
from molpipeline.mol2mol.scaffolds import MakeScaffoldGeneric, MurckoScaffold
from molpipeline.mol2mol.standardization import (
    ChargeParentExtractor,
    ExplicitHydrogenRemover,
    FragmentDeduplicator,
    IsotopeRemover,
    LargestFragmentChooser,
    MetalDisconnector,
    SaltRemover,
    SolventRemover,
    StereoRemover,
    TautomerCanonicalizer,
    Uncharger,
)

__all__ = (
    "TautomerCanonicalizer",
    "ChargeParentExtractor",
    "FragmentDeduplicator",
    "ElementFilter",
    "EmptyMoleculeFilter",
    "LargestFragmentChooser",
    "MakeScaffoldGeneric",
    "MetalDisconnector",
    "MixtureFilter",
    "MolToMolReaction",
    "MurckoScaffold",
    "ExplicitHydrogenRemover",
    "IsotopeRemover",
    "StereoRemover",
    "SaltRemover",
    "SolventRemover",
    "Uncharger",
    "InorganicsFilter",
    "SmartsFilter",
    "SmilesFilter",
    "RDKitDescriptorsFilter",
    "ComplexFilter",
)



"""Test smiles to mol pipeline element."""

import unittest
from typing import Any

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol


class TestSmiles2Mol(unittest.TestCase):
    """Test case for testing conversion of SMILES input to molecules."""

    def test_smiles2mol_explict_hydrogens(self) -> None:
        """Test smiles reading with and without explicit smiles."""
        smiles = "C[H]"

        # test: remove explicit Hs
        pipeline = Pipeline(
            [
                (
                    "Smiles2Mol",
                    SmilesToMol(remove_hydrogens=True),
                ),
            ]
        )
        mols = pipeline.fit_transform([smiles])
        self.assertEqual(len(mols), 1)
        self.assertIsNotNone(mols[0])
        self.assertEqual(mols[0].GetNumAtoms(), 1)

        # test: keep explicit Hs
        pipeline2 = Pipeline(
            [
                (
                    "Smiles2Mol",
                    SmilesToMol(remove_hydrogens=False),
                ),
            ]
        )
        mols2 = pipeline2.fit_transform([smiles])
        self.assertEqual(len(mols2), 1)
        self.assertIsNotNone(mols2[0])
        self.assertEqual(mols2[0].GetNumAtoms(), 2)

    def test_getter_setter(self) -> None:
        """Test getter and setter methods."""
        smiles2mol = SmilesToMol(remove_hydrogens=False)
        self.assertEqual(smiles2mol.get_params()["remove_hydrogens"], False)
        params: dict[str, Any] = {
            "remove_hydrogens": True,
        }
        smiles2mol.set_params(**params)
        self.assertEqual(smiles2mol.get_params()["remove_hydrogens"], True)



"""Contains functions for molecular similarity."""

from typing import Union

import numpy as np
import numpy.typing as npt
from scipy import sparse


def tanimoto_similarity_sparse(
    matrix_a: sparse.csr_matrix, matrix_b: sparse.csr_matrix
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto similarities between feature matrix a and b.

    Parameters
    ----------
    matrix_a: sparse.csr_matrix
        Feature matrix A.
    matrix_b: sparse.csr_matrix
        Feature matrix B.

    Returns
    -------
    npt.NDArray[np.float64]
        Matrix of similarity values between instances of A (rows/first dim) , and instances of B (columns/second dim).
    """
    intersection = matrix_a.dot(matrix_b.transpose()).toarray()
    norm_1 = np.array(matrix_a.multiply(matrix_a).sum(axis=1))
    norm_2 = np.array(matrix_b.multiply(matrix_b).sum(axis=1))
    union = norm_1 + norm_2.T - intersection
    # avoid division by zero https://stackoverflow.com/a/37977222
    return np.divide(
        intersection,
        union,
        out=np.zeros(intersection.shape, dtype=float),
        where=union != 0,
    )


def tanimoto_distance_sparse(
    matrix_a: sparse.csr_matrix, matrix_b: sparse.csr_matrix
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto distance between feature matrix a and b.

    Tanimoto distance is defined as 1-similarity.

    Parameters
    ----------
    matrix_a: sparse.csr_matrix
        Feature matrix A.
    matrix_b: sparse.csr_matrix
        Feature matrix B.

    Returns
    -------
    npt.NDArray[np.float64]
        Matrix of similarity values between instances of A (rows/first dim) , and instances of B (columns/second dim).
    """
    return 1 - tanimoto_similarity_sparse(matrix_a, matrix_b)  # type: ignore


def self_tanimoto_similarity(
    matrix_a: Union[sparse.csr_matrix, npt.NDArray[np.int_]],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto similarity between feature matrix a and itself.

    Parameters
    ----------
    matrix_a: Union[sparse.csr_matrix, npt.NDArray[np.int_]]
        Feature matrix.

    Returns
    -------
    npt.NDArray[np.float64]
        Square matrix of similarity values between all instances in the matrix.
    """
    if isinstance(matrix_a, np.ndarray):
        sparse_matrix = sparse.csr_matrix(matrix_a)
    elif isinstance(matrix_a, sparse.csr_matrix):
        sparse_matrix = matrix_a
    else:
        raise TypeError(f"Unsupported type: {type(matrix_a)}")
    return tanimoto_similarity_sparse(sparse_matrix, sparse_matrix)


def self_tanimoto_distance(
    matrix_a: Union[sparse.csr_matrix, npt.NDArray[np.int_]],
) -> npt.NDArray[np.float64]:
    """Calculate a matrix of tanimoto distance between feature matrix a and itself.

    Parameters
    ----------
    matrix_a: Union[sparse.csr_matrix, npt.NDArray[np.int_]]
        Feature matrix.

    Returns
    -------
    npt.NDArray[np.float64]
        Square matrix of similarity values between all instances in the matrix.
    """
    return 1 - self_tanimoto_similarity(matrix_a)  # type: ignore



"""Connected component clustering estimator."""

from __future__ import annotations

from numbers import Real
from typing import Any

import numpy as np
import numpy.typing as npt
from scipy import sparse
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, ClusterMixin, _fit_context
from sklearn.utils._param_validation import Interval
from sklearn.utils.validation import validate_data

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from molpipeline.estimators.algorithm.connected_component_clustering import (
    calc_chunk_size_from_memory_requirement,
    connected_components_iterative_algorithm,
)
from molpipeline.utils.kernel import tanimoto_similarity_sparse


class ConnectedComponentClustering(ClusterMixin, BaseEstimator):
    """Connected component clustering estimator."""

    _parameter_constraints: dict[str, Any] = {
        "distance_threshold": [Interval(Real, 0, None, closed="left")],
    }

    def __init__(
        self,
        distance_threshold: float,
        *,
        max_memory_usage: float = 4.0,
    ) -> None:
        """Initialize connected component clustering estimator.

        Parameters
        ----------
        distance_threshold : float
            Distance threshold for connected component clustering.
        max_memory_usage : float, optional
            Maximum memory usage in GB, by default 4.0 GB
        """
        self.distance_threshold: float = distance_threshold
        self.max_memory_usage: float = max_memory_usage
        self.n_clusters_: int | None = None
        self.labels_: npt.NDArray[np.int32] | None = None

    # pylint: disable=C0103,W0613
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: npt.NDArray[np.float64] | csr_matrix,
        y: npt.NDArray[np.float64] | None = None,
    ) -> Self:
        """Fit connected component clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Feature matrix.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        Self
            Fitted estimator.
        """
        X = validate_data(self, X=X, ensure_min_samples=2, accept_sparse=True)
        return self._fit(X)

    # pylint: disable=C0103,W0613
    def _fit(self, X: npt.NDArray[np.float64] | csr_matrix) -> Self:
        """Fit connected component clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Feature matrix.

        Returns
        -------
        Self
            Fitted estimator.
        """
        # convert tanimoto distance to similarity
        similarity_threshold: float = 1 - self.distance_threshold

        # get row chunk size based on 2D dense distance matrix that will be generated
        row_chunk_size = calc_chunk_size_from_memory_requirement(
            X.shape[0] * 2
            + 2,  # the self_tanimoto_distance needs two matrices of X.shape and two additional rows.
            X.shape[0],
            np.dtype("float64").itemsize,
            self.max_memory_usage,
        )

        if row_chunk_size >= X.shape[0]:
            similarity_matrix = tanimoto_similarity_sparse(X, X)
            adjacency_matrix = (similarity_matrix >= similarity_threshold).astype(
                np.int8
            )
            self.n_clusters_, self.labels_ = sparse.csgraph.connected_components(
                adjacency_matrix, directed=False, return_labels=True
            )
        else:
            self.n_clusters_, self.labels_ = connected_components_iterative_algorithm(
                X, similarity_threshold, row_chunk_size
            )
        return self

    def fit_predict(
        self,
        X: npt.NDArray[np.float64] | csr_matrix,  # pylint: disable=C0103
        y: npt.NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> npt.NDArray[np.int32]:
        """Fit and predict connected component clustering estimator.

        Parameters
        ----------
        X: npt.NDArray[np.float64] | csr_matrix
            Feature matrix of shape  (n_samples, n_features).
        y: Ignored
            Not used, present for API consistency by convention.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        np.ndarray[int]
            Cluster labels.
        """
        # pylint: disable=W0246
        return super().fit_predict(X, y, **kwargs)



"""Test the module post_prediction.py."""

import unittest

import numpy as np
from sklearn.base import clone
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier

from molpipeline.post_prediction import PostPredictionWrapper


class TestPostPredictionWrapper(unittest.TestCase):
    """Test the PostPredictionWrapper class."""

    def test_get_params(self) -> None:
        """Test get_params method."""
        rf = RandomForestClassifier()
        rf_params = rf.get_params(deep=True)

        ppw = PostPredictionWrapper(rf)
        ppw_params = ppw.get_params(deep=True)

        wrapped_params = {}
        for key, value in ppw_params.items():
            first, _, rest = key.partition("__")
            if first == "wrapped_estimator":
                if rest == "":
                    self.assertIs(rf, value)
                else:
                    wrapped_params[rest] = value

        self.assertDictEqual(rf_params, wrapped_params)

    def test_set_params(self) -> None:
        """Test set_params method."""
        rf = RandomForestClassifier()
        ppw = PostPredictionWrapper(rf)

        ppw.set_params(wrapped_estimator__n_estimators=10)
        self.assertIsInstance(ppw.wrapped_estimator, RandomForestClassifier)
        if not isinstance(ppw.wrapped_estimator, RandomForestClassifier):
            raise TypeError("Wrapped estimator is not a RandomForestClassifier.")
        self.assertEqual(ppw.wrapped_estimator.n_estimators, 10)

        ppw_params = ppw.get_params(deep=True)
        self.assertEqual(ppw_params["wrapped_estimator__n_estimators"], 10)

    def test_fit_transform(self) -> None:
        """Test fit method."""
        rng = np.random.default_rng(20240918)
        features = rng.random((10, 5))

        pca = PCA(n_components=3)
        pca.fit(features)
        pca_transformed = pca.transform(features)

        ppw = PostPredictionWrapper(clone(pca))
        ppw.fit(features)
        ppw_transformed = ppw.transform(features)

        self.assertEqual(pca_transformed.shape, ppw_transformed.shape)
        self.assertTrue(np.allclose(pca_transformed, ppw_transformed))

    def test_inverse_transform(self) -> None:
        """Test inverse_transform method."""
        rng = np.random.default_rng(20240918)
        features = rng.random((10, 5))

        pca = PCA(n_components=3)
        pca.fit(features)
        pca_transformed = pca.transform(features)
        pca_inverse = pca.inverse_transform(pca_transformed)

        ppw = PostPredictionWrapper(clone(pca))
        ppw.fit(features)
        ppw_transformed = ppw.transform(features)
        ppw_inverse = ppw.inverse_transform(ppw_transformed)

        self.assertEqual(features.shape, ppw_inverse.shape)
        self.assertEqual(pca_inverse.shape, ppw_inverse.shape)

        self.assertTrue(np.allclose(pca_inverse, ppw_inverse))



"""Test mol to bool conversion."""

import unittest

from molpipeline import Pipeline
from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.any2mol import AutoToMol
from molpipeline.mol2any import MolToBool


class TestMolToBool(unittest.TestCase):
    """Unittest for MolToBool."""

    def test_bool_conversion(self) -> None:
        """Test if the invalid instances are converted to bool."""
        mol2bool = MolToBool()
        result = mol2bool.transform(
            [
                1,
                2,
                InvalidInstance(element_id="test", message="test", element_name="Test"),
                4,
            ]
        )
        self.assertEqual(result, [True, True, False, True])

    def test_bool_conversion_pipeline(self) -> None:
        """Test if the invalid instances are converted to bool in pipeline."""
        pipeline = Pipeline(
            [
                ("auto_to_mol", AutoToMol()),
                ("mol2bool", MolToBool()),
            ]
        )
        result = pipeline.transform(["CC", "CCC", "no%valid~smiles"])
        self.assertEqual(result, [True, True, False])



"""Classes for handling substructures.

This is only relevant for explainable AI, where atoms need to be mapped to features.
"""

from __future__ import annotations

from rdkit import Chem


# pylint: disable=R0903
class AtomEnvironment:
    """A Class to store environment-information for fingerprint features."""

    def __init__(self, environment_atoms: set[int]):
        """Initialize AtomEnvironment.

        Parameters
        ----------
        environment_atoms: set[int]
            Indices of atoms encoded by environment.
        """
        self.environment_atoms = environment_atoms  # set of all atoms within radius


# pylint: disable=R0903
class CircularAtomEnvironment(AtomEnvironment):
    """A Class to store environment-information for morgan-fingerprint features."""

    def __init__(self, central_atom: int, radius: int, environment_atoms: set[int]):
        """Initialize CircularAtomEnvironment.

        Parameters
        ----------
        central_atom: int
            Index of central atom in circular fingerprint.
        radius: int
            Radius of feature.
        environment_atoms: set[int]
            All indices of atoms within radius of central atom.
        """
        super().__init__(environment_atoms)
        self.central_atom = central_atom
        self.radius = radius

    @classmethod
    def from_mol(
        cls, mol: Chem.Mol, central_atom_index: int, radius: int
    ) -> CircularAtomEnvironment:
        """Generate class from mol, using location (central_atom_index) and the radius.

        Parameters
        ----------
        mol: Chem.Mol
            Molecule from which the environment is derived.
        central_atom_index: int
            Index of central atom in feature.
        radius: int
            Radius of feature.

        Returns
        -------
        CircularAtomEnvironment
            Encoded the atoms which are within the radius of the central atom and are part of the feature.
        """
        if radius == 0:
            return CircularAtomEnvironment(
                central_atom_index, radius, {central_atom_index}
            )

        env = Chem.FindAtomEnvironmentOfRadiusN(mol, radius, central_atom_index)
        amap: dict[int, int] = {}
        _ = Chem.PathToSubmol(mol, env, atomMap=amap)
        env_atoms = amap.keys()
        return CircularAtomEnvironment(central_atom_index, radius, set(env_atoms))



"""Test the ignore error scorer wrapper."""

import unittest

import numpy as np
from sklearn import linear_model
from sklearn.metrics import get_scorer

from molpipeline.metrics import ignored_value_scorer


class IgnoreErrorScorerTest(unittest.TestCase):
    """Test the ignore error scorer wrapper."""

    def test_filter_nan(self) -> None:
        """Test that filtering np.nan works."""
        y_true = np.array([1, 0, 0, 1, 0])
        y_pred = np.array([1, 0, 0, 1, np.nan])
        ba_score = ignored_value_scorer("balanced_accuracy", np.nan)
        value = ba_score._score_func(y_true, y_pred)  # pylint: disable=protected-access
        self.assertAlmostEqual(value, 1.0)

    def test_filter_none(self) -> None:
        """Test that filtering None works."""
        y_true = np.array([1, 0, 0, 1, 0])
        y_pred = np.array([1, 0, 0, 1, None])
        ba_score = ignored_value_scorer("balanced_accuracy", None)
        value = ba_score._score_func(y_true, y_pred)  # pylint: disable=protected-access
        self.assertAlmostEqual(value, 1.0)

    def test_filter_nan_with_none(self) -> None:
        """Test that filtering NaN with None works."""
        y_true = np.array([1, 0, 0, 1, 0])
        y_pred = np.array([1, 0, 0, 1, None])
        ba_score = ignored_value_scorer("balanced_accuracy", np.nan)
        self.assertAlmostEqual(
            ba_score._score_func(y_true, y_pred),  # pylint: disable=protected-access
            1.0,
        )

    def test_filter_none_with_nan(self) -> None:
        """Test that filtering None with NaN works."""
        y_true = np.array([1, 0, 0, 1, 0])
        y_pred = np.array([1, 0, 0, 1, np.nan])
        ba_score = ignored_value_scorer("balanced_accuracy", None)
        self.assertAlmostEqual(
            ba_score._score_func(y_true, y_pred),  # pylint: disable=protected-access
            1.0,
        )

    def test_correct_init_mse(self) -> None:
        """Test that initialization is correct as we access via protected vars."""
        x_train = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]).reshape(
            -1, 1
        )
        y_train = np.array([0.1, 0.3, 0.3, 0.4, 0.5, 0.5, 0.7, 0.88, 0.9, 1])
        regr = linear_model.LinearRegression()
        regr.fit(x_train, y_train)
        cix_scorer = ignored_value_scorer("neg_mean_squared_error", None)
        scikit_scorer = get_scorer("neg_mean_squared_error")
        self.assertEqual(
            cix_scorer(regr, x_train, y_train), scikit_scorer(regr, x_train, y_train)
        )

    def test_correct_init_rmse(self) -> None:
        """Test that initialization is correct as we access via protected vars."""
        x_train = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]).reshape(
            -1, 1
        )
        y_train = np.array([0.1, 0.3, 0.3, 0.4, 0.5, 0.5, 0.7, 0.88, 0.9, 1])
        regr = linear_model.LinearRegression()
        regr.fit(x_train, y_train)
        cix_scorer = ignored_value_scorer("neg_root_mean_squared_error", None)
        scikit_scorer = get_scorer("neg_root_mean_squared_error")
        self.assertEqual(
            cix_scorer(regr, x_train, y_train), scikit_scorer(regr, x_train, y_train)
        )

    def test_correct_init_inheritance(self) -> None:
        """Test that initialization is correct if we pass an initialized scorer."""
        x_train = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]).reshape(
            -1, 1
        )
        y_train = np.array([0.1, 0.3, 0.3, 0.4, 0.5, 0.5, 0.7, 0.88, 0.9, 1])
        regr = linear_model.LinearRegression()
        regr.fit(x_train, y_train)
        scikit_scorer = get_scorer("neg_root_mean_squared_error")
        cix_scorer = ignored_value_scorer(
            get_scorer("neg_root_mean_squared_error"), None
        )
        self.assertEqual(
            cix_scorer(regr, x_train, y_train), scikit_scorer(regr, x_train, y_train)
        )



"""Classes for transforming rdkit molecules to any type of output."""

from __future__ import annotations

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.mol2any.mol2string import (
    MolToStringPipelineElement as _MolToStringPipelineElement,
)


class MolToSmiles(_MolToStringPipelineElement):
    """PipelineElement to transform a molecule to a SMILES string."""

    def pretransform_single(self, value: Chem.Mol) -> str:
        """Transform a molecule to a SMILES string.

        Parameters
        ----------
        value: Chem.Mol
            Molecule to be transformed to SMILES string.

        Returns
        -------
        str
            SMILES string of molecule.
        """
        return str(Chem.MolToSmiles(value))



"""Functions for creating default chemprop models."""

from typing import Any

from molpipeline.estimators.chemprop import ChempropModel, ChempropNeuralFP
from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BinaryClassificationFFN,
    BondMessagePassing,
    SumAggregation,
)


def get_binary_classification_mpnn() -> MPNN:
    """Get a Chemprop model for binary classification.

    Returns
    -------
    ChempropModel
        The Chemprop model.
    """
    binary_clf_ffn = BinaryClassificationFFN()
    aggregate = SumAggregation()
    bond_message_passing = BondMessagePassing()
    mpnn = MPNN(
        message_passing=bond_message_passing,
        agg=aggregate,
        predictor=binary_clf_ffn,
    )
    return mpnn


def get_neural_fp_encoder(
    init_kwargs: dict[str, Any] | None = None,
) -> ChempropNeuralFP:
    """Get the Chemprop model.

    Parameters
    ----------
    init_kwargs : dict[str, Any], optional
        Additional keyword arguments to pass to `ChempropNeuralFP` during initialization.

    Returns
    -------
    ChempropNeuralFP
        The Chemprop model.
    """
    mpnn = get_binary_classification_mpnn()
    init_kwargs = init_kwargs or {}
    chemprop_model = ChempropNeuralFP(
        model=mpnn, lightning_trainer__accelerator="cpu", **init_kwargs
    )
    return chemprop_model


def get_chemprop_model_binary_classification_mpnn() -> ChempropModel:
    """Get the Chemprop model.

    Returns
    -------
    ChempropModel
        The Chemprop model.
    """
    mpnn = get_binary_classification_mpnn()
    chemprop_model = ChempropModel(model=mpnn, lightning_trainer__accelerator="cpu")
    return chemprop_model



"""This module contains the default models used for testing molpipeline functions and classes."""

from sklearn.ensemble import RandomForestClassifier

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.mol2any import (
    MolToConcatenatedVector,
    MolToMorganFP,
    MolToRDKitPhysChem,
    MolToSmiles,
)
from molpipeline.mol2mol import (
    EmptyMoleculeFilter,
    FragmentDeduplicator,
    MetalDisconnector,
    MixtureFilter,
    SaltRemover,
    StereoRemover,
    TautomerCanonicalizer,
    Uncharger,
)
from molpipeline.mol2mol.filter import ElementFilter
from molpipeline.post_prediction import PostPredictionWrapper


def get_morgan_physchem_rf_pipeline(n_jobs: int = 1) -> Pipeline:
    """Get a pipeline combining Morgan fingerprints and physicochemical properties with a RandomForestClassifier.

    Parameters
    ----------
    n_jobs: int, default=-1
        Number of parallel jobs to use.

    Returns
    -------
    Pipeline
        A pipeline combining Morgan fingerprints and physicochemical properties with a RandomForestClassifier.
    """
    error_filter = ErrorFilter(filter_everything=True)
    pipeline = Pipeline(
        [
            ("smi2mol", SmilesToMol()),
            (
                "mol2fp",
                MolToConcatenatedVector(
                    [
                        ("morgan", MolToMorganFP(n_bits=2048)),
                        ("physchem", MolToRDKitPhysChem()),
                    ]
                ),
            ),
            ("error_filter", error_filter),
            ("rf", RandomForestClassifier(n_jobs=n_jobs)),
            (
                "filter_reinserter",
                PostPredictionWrapper(
                    FilterReinserter.from_error_filter(error_filter, None)
                ),
            ),
        ],
        n_jobs=n_jobs,
    )
    return pipeline


def get_standardization_pipeline(n_jobs: int = 1) -> Pipeline:
    """Get the standardization pipeline.

    Parameters
    ----------
    n_jobs: int, optional (default=-1)
        The number of jobs to use for standardization.
        In case of -1, all available CPUs are used.

    Returns
    -------
    Pipeline
        The standardization pipeline.
    """
    error_filter = ErrorFilter(filter_everything=True)
    # Set up pipeline
    standardization_pipeline = Pipeline(
        [
            ("smi2mol", SmilesToMol()),
            ("metal_disconnector", MetalDisconnector()),
            ("salt_remover", SaltRemover()),
            ("element_filter", ElementFilter()),
            ("uncharge1", Uncharger()),
            ("canonical_tautomer", TautomerCanonicalizer()),
            ("uncharge2", Uncharger()),
            ("stereo_remover", StereoRemover()),
            ("fragment_deduplicator", FragmentDeduplicator()),
            ("mixture_remover", MixtureFilter()),
            ("empty_molecule_remover", EmptyMoleculeFilter()),
            ("mol2smi", MolToSmiles()),
            ("error_filter", error_filter),
            ("error_replacer", FilterReinserter.from_error_filter(error_filter, None)),
        ],
        n_jobs=n_jobs,
    )
    return standardization_pipeline



"""Test Chemprop component wrapper."""

import unittest

from chemprop.nn.loss import LossFunction
from sklearn.base import clone
from torch import nn

from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BinaryClassificationFFN,
    BondMessagePassing,
    MeanAggregation,
    SumAggregation,
)


class BinaryClassificationFFNTest(unittest.TestCase):
    """Test the BinaryClassificationFFN class."""

    def test_get_set_params(self) -> None:
        """Test the get_params and set_params methods."""
        binary_clf_ffn = BinaryClassificationFFN()
        orig_params = binary_clf_ffn.get_params(deep=True)
        new_params = {
            "activation": "relu",
            "dropout": 0.5,
            "hidden_dim": 400,
            "input_dim": 300,
            "n_layers": 2,
            "n_tasks": 1,
        }
        # Check setting new parameters
        binary_clf_ffn.set_params(**new_params)
        model_params = binary_clf_ffn.get_params(deep=True)
        for param_name, param in new_params.items():
            self.assertEqual(param, model_params[param_name])

        # Check setting original parameters
        binary_clf_ffn.set_params(**orig_params)
        model_params = binary_clf_ffn.get_params(deep=True)
        for param_name, param in orig_params.items():
            self.assertEqual(param, model_params[param_name])


class BondMessagePassingTest(unittest.TestCase):
    """Test the BondMessagePassing class."""

    def test_get_set_params(self) -> None:
        """Test the get_params and set_params methods."""
        bond_message_passing = BondMessagePassing()
        orig_params = bond_message_passing.get_params(deep=True)
        new_params = {
            "activation": "relu",
            "bias": True,
            "d_e": 14,
            "d_h": 300,
            "d_v": 133,
            "d_vd": None,
            "depth": 4,
            "dropout_rate": 0.5,
            "undirected": False,
        }
        # Check setting new parameters
        bond_message_passing.set_params(**new_params)
        model_params = bond_message_passing.get_params(deep=True)
        for param_name, param in new_params.items():
            self.assertEqual(param, model_params[param_name])

        # Check setting original parameters
        bond_message_passing.set_params(**orig_params)
        model_params = bond_message_passing.get_params(deep=True)
        for param_name, param in orig_params.items():
            self.assertEqual(param, model_params[param_name])


class MPNNTest(unittest.TestCase):
    """Test the MPNN class."""

    def test_get_set_params(self) -> None:
        """Test the get_params and set_params methods."""
        mpnn1 = MPNN(
            message_passing=BondMessagePassing(depth=2),
            agg=SumAggregation(),
            predictor=BinaryClassificationFFN(n_layers=1),
        )
        params1 = mpnn1.get_params(deep=True)

        mpnn2 = MPNN(
            message_passing=BondMessagePassing(depth=1),
            agg=MeanAggregation(),
            predictor=BinaryClassificationFFN(n_layers=4),
        )
        mpnn2.set_params(**params1)
        for param_name, param in mpnn1.get_params(deep=True).items():
            param2 = mpnn2.get_params(deep=True)[param_name]
            # Classes are cloned, so they are not equal, but they should be the same class
            # Since (here) objects are identical if their parameters are identical, and since all
            # their parameters are listed flat in the params dicts, all objects are identical if
            # param dicts are identical.
            if hasattr(param, "get_params"):
                self.assertEqual(param.__class__, param2.__class__)
            else:
                self.assertEqual(param, param2)

    def test_clone(self) -> None:
        """Test the clone method."""
        mpnn = MPNN(
            message_passing=BondMessagePassing(),
            agg=SumAggregation(),
            predictor=BinaryClassificationFFN(),
        )
        mpnn_clone = clone(mpnn)
        for param_name, param in mpnn.get_params(deep=True).items():
            clone_param = mpnn_clone.get_params(deep=True)[param_name]
            if hasattr(param, "get_params"):
                self.assertEqual(param.__class__, clone_param.__class__)
            elif isinstance(param, LossFunction):
                self.assertEqual(
                    param.state_dict()["task_weights"],
                    clone_param.state_dict()["task_weights"],
                )
                self.assertEqual(type(param), type(clone_param))
            elif isinstance(param, nn.Identity):
                self.assertEqual(type(param), type(clone_param))
            else:
                self.assertEqual(param, clone_param)


if __name__ == "__main__":
    unittest.main()



"""Functions of fingerprints for comparing output with molpipline."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from rdkit import Chem

# pylint: disable=no-name-in-module
from rdkit.Chem import rdFingerprintGenerator as rdkit_fp
from rdkit.DataStructs import ExplicitBitVect, UIntSparseIntVect
from scipy import sparse


def make_sparse_fp(
    smiles_list: list[str], radius: int, n_bits: int
) -> sparse.csr_matrix:
    """Create a sparse Morgan fingerprint matrix from a list of SMILES.

    Used in Unittests.

    Parameters
    ----------
    smiles_list: list[str]
        SMILES representations of molecules which will be encoded as fingerprint.
    radius: int
        Radius of features.
    n_bits: int
        Obtained features will be mapped to a vector of size n_bits.

    Returns
    -------
    sparse.csr_matrix
        Feature matrix.
    """
    vector_list = []
    morgan_fp = rdkit_fp.GetMorganGenerator(radius=radius, fpSize=n_bits)
    for smiles in smiles_list:
        mol = Chem.MolFromSmiles(smiles)  # pylint: disable=no-member
        vector = morgan_fp.GetFingerprintAsNumPy(mol)
        vector_list.append(sparse.csr_matrix(vector))
    return sparse.vstack(vector_list)


def fingerprints_to_numpy(
    fingerprints: list[ExplicitBitVect] | sparse.csr_matrix | npt.NDArray[np.int_],
) -> npt.NDArray[np.int_]:
    """Convert fingerprints in various types to numpy.

    Parameters
    ----------
    fingerprints: list[ExplicitBitVect] | sparse.csr_matrix | npt.NDArray[np.int_]
        Fingerprint matrix.

    Returns
    -------
    npt.NDArray
        Numpy fingerprint matrix.
    """
    if all(isinstance(fp, ExplicitBitVect) for fp in fingerprints):
        return np.array(fingerprints)
    if all(isinstance(fp, UIntSparseIntVect) for fp in fingerprints):
        return np.array([fp.ToList() for fp in fingerprints])
    if isinstance(fingerprints, sparse.csr_matrix):
        return fingerprints.toarray()
    if isinstance(fingerprints, np.ndarray):
        return fingerprints
    raise ValueError("Unknown fingerprint type. Can not convert to numpy")



"""Module for utilities converting values."""

from typing import Sequence

from molpipeline.utils.molpipeline_types import IntCountRange, IntOrIntCountRange


def count_value_to_tuple(count: IntOrIntCountRange) -> IntCountRange:
    """Convert a count value to a tuple.

    Parameters
    ----------
    count: Union[int, IntCountRange]
        Count value. Can be a single int or a tuple of two values.

    Returns
    -------
    IntCountRange
        Tuple of count values.
    """
    if isinstance(count, int):
        return count, count
    if isinstance(count, Sequence):
        count_tuple = tuple(count)
        if len(count_tuple) != 2:
            raise ValueError(f"Expected a sequence of length 2, got: {count_tuple}")
        return count_tuple
    raise TypeError(f"Got unexpected type: {type(count)}")



"""Test connected component clustering estimator."""

from __future__ import annotations

import unittest

import numpy as np
from scipy.sparse import csr_matrix

from molpipeline.estimators import ConnectedComponentClustering


class TestConnectedComponentClusteringEstimator(unittest.TestCase):
    """Test connected component clustering estimator."""

    def test_connected_component_clustering_estimator(self) -> None:
        """Test connected component clustering estimator."""
        ccc = ConnectedComponentClustering(distance_threshold=0.5, max_memory_usage=0.1)

        # test no chunking needed
        self.assertEqual(
            ccc.fit_predict(csr_matrix([[1, 0, 1], [0, 1, 0], [1, 0, 1]])).tolist(),
            [0, 1, 0],
        )
        self.assertEqual(ccc.n_clusters_, 2)

        # test chunking needed
        matrix = csr_matrix([[1, 0, 1], [0, 1, 0], [1, 0, 1]])
        nof_bytes_per_row = np.dtype("float64").itemsize * matrix.shape[1] / (1 << 30)
        ccc = ConnectedComponentClustering(
            distance_threshold=0.5,
            max_memory_usage=nof_bytes_per_row,
        )
        self.assertEqual(
            ccc.fit_predict(matrix).tolist(),
            [0, 1, 0],
        )
        self.assertEqual(ccc.n_clusters_, 2)



"""Abstract classes for creating rdkit molecules from string representations."""

from __future__ import annotations

import abc

from molpipeline.abstract_pipeline_elements.core import (
    AnyToMolPipelineElement,
    InvalidInstance,
)
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol


class StringToMolPipelineElement(AnyToMolPipelineElement, abc.ABC):
    """Abstract class for PipelineElements which transform molecules to integer vectors."""

    _input_type = "str"
    _output_type = "RDKitMol"

    def transform(self, values: list[str]) -> list[OptionalMol]:
        """Transform the list of molecules to sparse matrix.

        Parameters
        ----------
        values: list[str]
            List of string representations of molecules which are transformed to RDKit molecules.

        Returns
        -------
        list[OptionalMol]
            List of RDKit molecules. If a string representation could not be transformed to a molecule, None is returned.
        """
        return super().transform(values)

    @abc.abstractmethod
    def pretransform_single(self, value: str) -> OptionalMol:
        """Transform mol to a string.

        Parameters
        ----------
        value: str
            Representation transformed to a RDKit molecule.

        Returns
        -------
        OptionalMol
            RDKit molecule if representation was valid, else InvalidInstance.
        """


class SimpleStringToMolElement(StringToMolPipelineElement, abc.ABC):
    """Transforms string representation to RDKit Mol objects."""

    def pretransform_single(self, value: str) -> OptionalMol:
        """Transform string to molecule.

        Parameters
        ----------
        value: str
            string representation.

        Returns
        -------
        OptionalMol
            Rdkit molecule if valid string representation, else None.
        """
        if value is None:
            return InvalidInstance(
                self.uuid,
                f"Invalid representation: {value}",
                self.name,
            )

        if not isinstance(value, str):
            return InvalidInstance(
                self.uuid,
                f"Not a string: {value}",
                self.name,
            )

        mol: RDKitMol = self.string_to_mol(value)

        if not mol:
            return InvalidInstance(
                self.uuid,
                f"Invalid representation: {value}",
                self.name,
            )
        mol.SetProp("identifier", value)
        return mol

    @abc.abstractmethod
    def string_to_mol(self, value: str) -> RDKitMol:
        """Transform string representation to molecule.

        Parameters
        ----------
        value: str
            string representation

        Returns
        -------
        RDKitMol
            Rdkit molecule if valid representation, else None.
        """



"""Module for generating heatmaps from 2D-grids.

Much of the visualization code in this file originates from projects of Christian W. Feldmann:
    https://github.com/c-feldmann/rdkit_heatmaps
    https://github.com/c-feldmann/compchemkit
"""

import abc
from typing import Callable, Sequence

import numpy as np
import numpy.typing as npt
from matplotlib import colors
from rdkit.Chem import Draw
from rdkit.Geometry.rdGeometry import Point2D


class Grid2D(abc.ABC):
    """Metaclass for discrete 2-dimensional grids.

    This class holds a matrix of values accessed by index, where each cell is associated with a specific location.
    """

    def __init__(
        self,
        x_lim: Sequence[float],
        y_lim: Sequence[float],
        x_res: int,
        y_res: int,
    ) -> None:
        """Initialize the Grid2D with limits and resolution of the axes.

        Parameters
        ----------
        x_lim: Sequence[float]
            Extend of the grid along the x-axis (xmin, xmax).
        y_lim: Sequence[float]
            Extend of the grid along the y-axis (ymin, ymax).
        x_res: int
            Resolution (number of cells) along x-axis.
        y_res: int
            Resolution (number of cells) along y-axis.
        """
        if len(x_lim) != 2:
            raise ValueError("x_lim must be of length 2.")
        if len(y_lim) != 2:
            raise ValueError("y_lim must be of length 2.")

        self.x_lim = x_lim
        self.y_lim = y_lim
        self.x_res = x_res
        self.y_res = y_res
        self._dx = (max(self.x_lim) - min(self.x_lim)) / self.x_res
        self._dy = (max(self.y_lim) - min(self.y_lim)) / self.y_res
        self.values = np.zeros((self.x_res, self.y_res))

    @property
    def dx(self) -> float:
        """Length of cell in x-direction."""
        return self._dx

    @property
    def dy(self) -> float:
        """Length of cell in y-direction."""
        return self._dy

    def grid_field_center(self, x_idx: int, y_idx: int) -> tuple[float, float]:
        """Center of cell specified by index along x and y.

        Parameters
        ----------
        x_idx: int
            cell-index along x-axis.
        y_idx: int
            cell-index along y-axis.

        Returns
        -------
        tuple[float, float]
            Coordinates of center of cell.
        """
        x_coord = min(self.x_lim) + self.dx * (x_idx + 0.5)
        y_coord = min(self.y_lim) + self.dy * (y_idx + 0.5)
        return x_coord, y_coord

    def grid_field_lim(
        self, x_idx: int, y_idx: int
    ) -> tuple[tuple[float, float], tuple[float, float]]:
        """Get x and y coordinates for the upper left and lower right position of specified pixel.

        Parameters
        ----------
        x_idx: int
            cell-index along x-axis.
        y_idx: int
            cell-index along y-axis.

        Returns
        -------
        tuple[tuple[float, float], tuple[float, float]]
            Coordinates of upper left and lower right corner of cell.
        """
        upper_left = (
            min(self.x_lim) + self.dx * x_idx,
            min(self.y_lim) + self.dy * y_idx,
        )
        lower_right = (
            min(self.x_lim) + self.dx * (x_idx + 1),
            min(self.y_lim) + self.dy * (y_idx + 1),
        )
        return upper_left, lower_right


class ColorGrid(Grid2D):
    """Stores rgba-values of cells."""

    def __init__(
        self,
        x_lim: Sequence[float],
        y_lim: Sequence[float],
        x_res: int,
        y_res: int,
    ):
        """Initialize the ColorGrid with limits and resolution of the axes.

        Parameters
        ----------
        x_lim: Sequence[float]
            Extend of the grid along the x-axis (xmin, xmax).
        y_lim: Sequence[float]
            Extend of the grid along the y-axis (ymin, ymax).
        x_res: int
            Resolution (number of cells) along x-axis.
        y_res: int
            Resolution (number of cells) along y-axis.
        """
        super().__init__(x_lim, y_lim, x_res, y_res)
        self.color_grid = np.ones((self.x_res, self.y_res, 4))


class ValueGrid(Grid2D):
    """Calculate and store values of cells.

    Evaluates all added functions for the position of each cell and calculates the value of each cell as sum of these
    functions.
    """

    # list of functions to be evaluated for each grid cell
    function_list: list[Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]]

    def __init__(
        self,
        x_lim: Sequence[float],
        y_lim: Sequence[float],
        x_res: int,
        y_res: int,
        function_list: (
            list[Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]] | None
        ) = None,
    ):
        """Initialize the ValueGrid with limits and resolution of the axes.

        Parameters
        ----------
        x_lim: Sequence[float]
            Extend of the grid along the x-axis (xmin, xmax).
        y_lim: Sequence[float]
            Extend of the grid along the y-axis (ymin, ymax).
        x_res: int
            Resolution (number of cells) along x-axis.
        y_res: int
            Resolution (number of cells) along y-axis.
        function_list: list[Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]], optional
            List of functions to be evaluated for each cell, by default None.
        """
        super().__init__(x_lim, y_lim, x_res, y_res)
        if function_list is not None:
            self.function_list = function_list
        else:
            self.function_list = []
        self.values = np.zeros((self.x_res, self.y_res))

    def add_function(
        self, function: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]
    ) -> None:
        """Add a function to the grid which is evaluated for each cell, when `self.evaluate` is called.

        Parameters
        ----------
        function: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]
            Function to be evaluated for each cell. The function should take an array of positions and return an array
            of values, e.g. a Gaussian function.
        """
        self.function_list.append(function)

    def evaluate(self) -> None:
        """Evaluate each function for each cell. Values of cells are calculated as the sum of all function-values.

        The results are saved to `self.values`.
        """
        self.values = np.zeros((self.x_res, self.y_res))
        x_y0_list = np.array(
            [self.grid_field_center(x, 0)[0] for x in range(self.x_res)]
        )
        x0_y_list = np.array(
            [self.grid_field_center(0, y)[1] for y in range(self.y_res)]
        )
        xv, yv = np.meshgrid(x_y0_list, x0_y_list)
        xv = xv.ravel()
        yv = yv.ravel()
        coordinate_pairs = np.vstack([xv, yv]).T
        for f in self.function_list:
            values = f(coordinate_pairs)
            values = values.reshape(self.y_res, self.x_res).T
            if values.shape != self.values.shape:
                raise AssertionError(
                    f"Function does not return correct shape. Shape was {(values.shape, self.values.shape)}"
                )
            self.values += values

    def map2color(
        self,
        c_map: colors.Colormap,
        normalizer: colors.Normalize,
    ) -> ColorGrid:
        """Generate a ColorGrid from `self.values` according to given colormap.

        Parameters
        ----------
        c_map: colors.Colormap
            Colormap to be used for mapping values to colors.
        normalizer: colors.Normalize
            Normalizer to be used for mapping values to colors.

        Returns
        -------
        ColorGrid
            ColorGrid with colors corresponding to ValueGrid.
        """
        color_grid = ColorGrid(self.x_lim, self.y_lim, self.x_res, self.y_res)
        norm = normalizer(self.values)
        color_grid.color_grid = np.array(c_map(norm))
        return color_grid


def get_color_normalizer_from_data(
    values: npt.NDArray[np.float64],
) -> colors.Normalize:
    """Create a color normalizer based on the data distribution of 'values'.

    Parameters
    ----------
    values: npt.NDArray[np.float64]
        Data to derive limits of normalizer. The maximum absolute value of
        values` is used as limit.

    Returns
    -------
    colors.Normalize
        Normalizer for colors.
    """
    abs_max = np.max(np.abs(values))
    normalizer = colors.Normalize(vmin=-abs_max, vmax=abs_max)
    return normalizer


def color_canvas(canvas: Draw.MolDraw2D, color_grid: ColorGrid) -> None:
    """Draw a ColorGrid object to a RDKit Draw.MolDraw2D canvas.

    Each pixel is drawn as rectangle, so if you use Draw.MolDrawSVG brace yourself and your RAM!

    Parameters
    ----------
    canvas: Draw.MolDraw2D
        RDKit Draw.MolDraw2D canvas.
    color_grid: ColorGrid
        ColorGrid object to be drawn on the canvas.
    """
    # draw only grid points whose color is not white.
    # we check for the exact values of white (1,1,1). np.isclose returns almost the same pixels but is slightly slower.
    mask = np.where(~np.all(color_grid.color_grid[:, :, :3] == [1, 1, 1], axis=2))

    for x, y in zip(*mask):
        upper_left, lower_right = color_grid.grid_field_lim(x, y)
        upper_left, lower_right = Point2D(*upper_left), Point2D(*lower_right)
        canvas.SetColour(tuple(color_grid.color_grid[x, y]))
        canvas.DrawRect(upper_left, lower_right)



"""Test logging utils."""

import io
import unittest
from contextlib import redirect_stdout

from molpipeline.utils.logging import print_elapsed_time


class LoggingUtilsTest(unittest.TestCase):
    """Unittest for conversion of sklearn models to json and back."""

    def test__print_elapsed_time(self) -> None:
        """Test message logging with timings work as expected."""

        # when message is None nothing should be printed
        stream1 = io.StringIO()
        with redirect_stdout(stream1):
            with print_elapsed_time("source", message=None, use_logger=False):
                pass
        output1 = stream1.getvalue()
        self.assertEqual(output1, "")

        # message should be printed in the expected sklearn format
        stream2 = io.StringIO()
        with redirect_stdout(stream2):
            with print_elapsed_time("source", message="my message", use_logger=False):
                pass
        output2 = stream2.getvalue()
        self.assertTrue(
            output2.startswith(
                "[source] ................................... my message, total="
            )
        )
